# Do not import this file directly. Use
#
#   from dials.array_family import flex
#

from __future__ import annotations

import collections
import copy
import functools
import itertools
import logging
import operator
import os
import pickle
from typing import List, Tuple

import numpy as np
import pandas as pd
from annlib_ext import AnnAdaptorSelfInclude

import boost_adaptbx.boost.python
import cctbx.array_family.flex
import cctbx.miller
import libtbx.smart_open
from scitbx import matrix

import dials.extensions.glm_background_ext
import dials.extensions.simple_centroid_ext
import dials.util.ext
import dials_array_family_flex_ext
from dials.algorithms.centroid import centroid_px_to_mm_panel

__all__ = ["real", "reflection_table_selector"]

logger = logging.getLogger(__name__)

# Set the 'real' type to either float or double
if dials_array_family_flex_ext.get_real_type() == "float":
    real = cctbx.array_family.flex.float
elif dials_array_family_flex_ext.get_real_type() == "double":
    real = cctbx.array_family.flex.double
else:
    raise TypeError('unknown "real" type')


@boost_adaptbx.boost.python.inject_into(dials_array_family_flex_ext.reflection_table)
class _:
    """
    An injector class to add additional methods to the reflection table.
    """

    # Set the default algorithms. These are set as class variables so that if they
    # are changed in the class, all new instances of reflection table will have
    # the modified algorithms. If these are modified on the instance level, then
    # only the instance will have the modified algorithms and new instances will
    # have the defaults
    background_algorithm = functools.partial(
        dials.extensions.glm_background_ext.GLMBackgroundExt, None
    )
    centroid_algorithm = functools.partial(
        dials.extensions.simple_centroid_ext.SimpleCentroidExt, None
    )

    @staticmethod
    def from_predictions(
        experiment, dmin=None, dmax=None, margin=1, force_static=False, padding=0
    ):
        """
        Construct a reflection table from predictions.

        :param experiment: The experiment to predict from
        :param dmin: The maximum resolution
        :param dmax: The minimum resolution
        :param margin: The margin to predict around
        :param force_static: Do static prediction with a scan varying model
        :param padding: Padding in degrees
        :return: The reflection table of predictions
        """
        if experiment.profile is not None:
            return experiment.profile.predict_reflections(
                experiment.imageset,
                experiment.crystal,
                experiment.beam,
                experiment.detector,
                experiment.goniometer,
                experiment.scan,
                dmin=dmin,
                dmax=dmax,
                margin=margin,
                force_static=force_static,
                padding=padding,
            )
        from dials.algorithms.spot_prediction.reflection_predictor import (
            ReflectionPredictor,
        )

        predict = ReflectionPredictor(
            experiment,
            dmin=dmin,
            dmax=dmax,
            margin=margin,
            force_static=force_static,
            padding=padding,
        )
        return predict()

    @staticmethod
    def from_predictions_multi(
        experiments, dmin=None, dmax=None, margin=1, force_static=False, padding=0
    ):
        """
        Construct a reflection table from predictions.

        :param experiments: The experiment list to predict from
        :param dmin: The maximum resolution
        :param dmax: The minimum resolution
        :param margin: The margin to predict around
        :param force_static: Do static prediction with a scan varying model
        :param padding: Padding in degrees
        :return: The reflection table of predictions
        """
        result = dials_array_family_flex_ext.reflection_table()
        for i, e in enumerate(experiments):
            rlist = dials_array_family_flex_ext.reflection_table.from_predictions(
                e,
                dmin=dmin,
                dmax=dmax,
                margin=margin,
                force_static=force_static,
                padding=padding,
            )
            rlist["id"] = cctbx.array_family.flex.int(len(rlist), i)
            if e.identifier:
                rlist.experiment_identifiers()[i] = e.identifier
            result.extend(rlist)
        return result

    @staticmethod
    def from_observations(experiments, params=None, is_stills=False):
        """
        Construct a reflection table from observations.

        :param experiments: The experiments
        :param params: The input parameters
        :param is_stills:   [ADVANCED] Force still-handling of experiment
                            ID remapping for dials.stills_process. Do
                            not use for general processing unless you
                            know all the implications.
        :return: The reflection table of observations
        """
        from dials.algorithms.spot_finding.factory import SpotFinderFactory

        if params is None:
            from dials.command_line.find_spots import phil_scope
            from dials.util.phil import parse

            params = phil_scope.fetch(source=parse("")).extract()

        if params.spotfinder.filter.min_spot_size is libtbx.Auto:
            detector = experiments[0].imageset.get_detector()
            if detector[0].get_type() == "SENSOR_PAD":
                # smaller default value for pixel array detectors
                params.spotfinder.filter.min_spot_size = 3
            else:
                params.spotfinder.filter.min_spot_size = 6
            logger.info(
                "Setting spotfinder.filter.min_spot_size=%i",
                params.spotfinder.filter.min_spot_size,
            )

        # Get the integrator from the input parameters
        logger.info("Configuring spot finder from input parameters")
        spotfinder = SpotFinderFactory.from_parameters(
            experiments=experiments, params=params, is_stills=is_stills
        )

        # Find the spots
        return spotfinder.find_spots(experiments)

    @staticmethod
    def from_pickle(filename):
        """
        Read the reflection table from pickle file.

        :param filename: The pickle filename
        :return: The reflection table
        """
        if filename and hasattr(filename, "__fspath__"):
            filename = filename.__fspath__()
        with libtbx.smart_open.for_reading(filename, "rb") as infile:
            result = pickle.load(infile, encoding="bytes")
            assert isinstance(result, dials_array_family_flex_ext.reflection_table)
            return result

    def as_msgpack_file(self, filename):
        """
        Write the reflection table to file in msgpack format
        """
        if filename and hasattr(filename, "__fspath__"):
            filename = filename.__fspath__()
        with libtbx.smart_open.for_writing(filename, "wb") as outfile:
            self.as_msgpack_to_file(dials.util.ext.streambuf(python_file_obj=outfile))

    @staticmethod
    def from_msgpack_file(filename):
        """
        Read the reflection table from file in msgpack format
        """
        if filename and hasattr(filename, "__fspath__"):
            filename = filename.__fspath__()
        with libtbx.smart_open.for_reading(filename, "rb") as infile:
            return dials_array_family_flex_ext.reflection_table.from_msgpack(
                infile.read()
            )

    def as_file(self, filename):
        """
        Write the reflection table to file in either msgpack or pickle format
        """
        if os.getenv("DIALS_USE_PICKLE"):
            self.as_pickle(filename)
        else:
            self.as_msgpack_file(filename)

    @staticmethod
    def from_file(filename):
        """
        Read the reflection table from either pickle or msgpack
        """
        try:
            return dials_array_family_flex_ext.reflection_table.from_msgpack_file(
                filename
            )
        except RuntimeError:
            return dials_array_family_flex_ext.reflection_table.from_pickle(filename)

    @staticmethod
    def empty_standard(nrows):
        """
        Create an empty table of specified number of rows with most of the standard
        keys

        :param nrows: The number of rows to create
        :return: The reflection table
        """

        assert nrows > 0
        table = dials_array_family_flex_ext.reflection_table(nrows)

        # General properties
        table["flags"] = cctbx.array_family.flex.size_t(nrows, 0)
        table["id"] = cctbx.array_family.flex.int(nrows, 0)
        table["panel"] = cctbx.array_family.flex.size_t(nrows, 0)

        # Predicted properties
        table["miller_index"] = cctbx.array_family.flex.miller_index(nrows)
        table["entering"] = cctbx.array_family.flex.bool(nrows)
        table["s1"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        table["xyzcal.mm"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        table["xyzcal.px"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        # table['ub_matrix'] = cctbx.array_family.flex.mat3_double(nrows, (0, 0, 0, 0, 0, 0, 0, 0, 0))

        # Observed properties
        table["xyzobs.px.value"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        table["xyzobs.px.variance"] = cctbx.array_family.flex.vec3_double(
            nrows, (0, 0, 0)
        )
        table["xyzobs.mm.value"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        table["xyzobs.mm.variance"] = cctbx.array_family.flex.vec3_double(
            nrows, (0, 0, 0)
        )
        table["rlp"] = cctbx.array_family.flex.vec3_double(nrows, (0, 0, 0))
        table["intensity.sum.value"] = cctbx.array_family.flex.double(nrows, 0)
        table["intensity.sum.variance"] = cctbx.array_family.flex.double(nrows, 0)
        table["intensity.prf.value"] = cctbx.array_family.flex.double(nrows, 0)
        table["intensity.prf.variance"] = cctbx.array_family.flex.double(nrows, 0)
        table["lp"] = cctbx.array_family.flex.double(nrows, 0)
        table["profile.correlation"] = cctbx.array_family.flex.double(nrows, 0)

        return table

    @staticmethod
    def plot(table, detector, key):
        """
        Plot a reflection table using matplotlib

        :param table: The reflection table
        :param detector: The detector model
        :param key: The key to plot
        """
        from matplotlib import pyplot as plt
        from matplotlib.patches import Polygon

        fig = plt.figure()
        ax = fig.add_subplot(111, aspect="equal")
        spots = table[key]
        if "px" in key:
            spots = [
                detector[table["panel"][i]].get_pixel_lab_coord(spots[i][0:2])
                for i in range(len(spots))
            ]
        else:
            assert "mm" in key
            spots = [
                detector[table["panel"][i]].get_lab_coord(spots[i][0:2])
                for i in range(len(spots))
            ]

        min_f = max_f = min_s = max_s = 0

        for i, panel in enumerate(detector):
            fs, ss = panel.get_image_size()
            p0 = panel.get_pixel_lab_coord((0, 0))
            p1 = panel.get_pixel_lab_coord((fs - 1, 0))
            p2 = panel.get_pixel_lab_coord((fs - 1, ss - 1))
            p3 = panel.get_pixel_lab_coord((0, ss - 1))
            p = Polygon(
                (p0[0:2], p1[0:2], p2[0:2], p3[0:2]),
                closed=True,
                color="green",
                fill=False,
                hatch="/",
            )

            if p.xy[:, 0].min() < min_f:
                min_f = p.xy[:, 0].min()
            if p.xy[:, 0].max() > max_f:
                max_f = p.xy[:, 0].max()
            if p.xy[:, 1].min() < min_s:
                min_s = p.xy[:, 1].min()
            if p.xy[:, 1].max() > max_s:
                max_s = p.xy[:, 1].max()

            ax.add_patch(p)

        ax.set_xlim((min_f - 10, max_f + 10))
        ax.set_ylim((min_s - 10, max_s + 10))
        plt.scatter([s[0] for s in spots], [s[1] for s in spots], c="blue", linewidth=0)
        plt.show()

    def as_pickle(self, filename):
        """
        Write the reflection table as a pickle file.

        :param filename: The output filename
        """
        # Clean up any removed experiments from the identifiers map
        self.clean_experiment_identifiers_map()

        with libtbx.smart_open.for_writing(filename, "wb") as outfile:
            pickle.dump(self, outfile, protocol=pickle.HIGHEST_PROTOCOL)

    def as_h5(self, filename):
        """
        Write the reflection table as a HDF5 file.

        :param filename: The output filename
        """
        from dials.util.nexus_old import NexusFile

        handle = NexusFile(filename, "w")
        # Clean up any removed experiments from the identifiers map
        self.clean_experiment_identifiers_map()
        handle.set_reflections(self)
        handle.close()

    def as_miller_array(self, experiment, intensity="sum"):
        """Return a miller array with the chosen intensities.

        Use the provided experiment object and intensity choice to make a miller
        intensity array with sigmas (no scaling applied).

        Args:
            experiment (dxtbx.model.Experiment): An experiment object.
            intensity (str): The intensity type that will be used to make the
                miller array e.g 'prf', 'sum'.

        Returns:
            cctbx.miller.array: A miller array with intensities and sigmas.

        Raises:
            KeyError: If chosen intensity values cannot be found in the table.
        """

        try:
            intensities, variances = (
                self["intensity." + intensity + ".value"],
                self["intensity." + intensity + ".variance"],
            )
        except KeyError as e:
            logger.error(e, exc_info=True)
            raise KeyError(
                "Unable to find %s, %s in reflection table"
                % (
                    "intensity." + intensity + ".value",
                    "intensity." + intensity + ".variance",
                )
            )

        miller_set = cctbx.miller.set(
            crystal_symmetry=experiment.crystal.get_crystal_symmetry(),
            indices=self["miller_index"],
            anomalous_flag=False,
        )
        i_obs = cctbx.miller.array(miller_set, data=intensities)
        i_obs.set_observation_type_xray_intensity()
        i_obs.set_sigmas(cctbx.array_family.flex.sqrt(variances))
        i_obs.set_info(
            cctbx.miller.array_info(source="DIALS", source_type="reflection_tables")
        )
        return i_obs

    def copy(self):
        """
        Copy everything.

        :return: A copy of the reflection table
        """
        return self.select(cctbx.array_family.flex.bool(len(self), True))

    def sort(self, name, reverse=False, order=None):
        """
        Sort the reflection table by a key.

        :param name: The name of the column
        :param reverse: Reverse the sort order
        :param order: For multi element items specify order
        """

        if type(self[name]) in (
            cctbx.array_family.flex.vec2_double,
            cctbx.array_family.flex.vec3_double,
            cctbx.array_family.flex.mat3_double,
            dials_array_family_flex_ext.int6,
            cctbx.array_family.flex.miller_index,
        ):
            data = self[name]
            if not order:
                perm = cctbx.array_family.flex.size_t(
                    sorted(range(len(self)), key=lambda x: data[x], reverse=reverse)
                )
            else:
                assert len(order) == len(data[0])
                perm = cctbx.array_family.flex.size_t(
                    sorted(
                        range(len(self)),
                        key=lambda x: tuple(data[x][i] for i in order),
                        reverse=reverse,
                    )
                )
        else:
            perm = cctbx.array_family.flex.sort_permutation(
                self[name], reverse=reverse, stable=True
            )
        self.reorder(perm)

    """
    Sorting the reflection table within an already sorted column
    """

    def subsort(self, key0, key1, reverse=False):
        """
        Sort the reflection based on key1 within a constant key0.

        :param key0: The name of the column values to sort within
        :param key1: The sorting key name within the selected column
        """
        uniq_values = self[key0]
        for ii in set(uniq_values):
            val = (uniq_values == ii).iselection()
            ref_tmp = copy.deepcopy(self[min(val) : (max(val) + 1)])
            ref_tmp.sort(key1, reverse)
            self[min(val) : (max(val) + 1)] = ref_tmp

    def match_by_hkle(
        self, other: dials_array_family_flex_ext.reflection_table
    ) -> Tuple[cctbx.array_family.flex.size_t, cctbx.array_family.flex.size_t]:
        """
        Match reflections with another set of reflections by the h, k, l
        and entering values. Uses pandas dataframe merge method to match
        the columns: assumes the key h, k, l, e is unique which is false
        if > 360 degree rotation.

        Args:
            other: reflection table to match against

        Returns:
            Indices in self, indices in other for matches
        """

        hkl = self["miller_index"].as_vec3_double().parts()
        hkl = (part.as_numpy_array().astype(int) for part in hkl)
        e = self["entering"].as_numpy_array()
        n = np.arange(e.size)
        p0 = pd.DataFrame(dict(zip("hklen", (*hkl, e, n))), copy=False)

        hkl = other["miller_index"].as_vec3_double().parts()
        hkl = (part.as_numpy_array().astype(int) for part in hkl)
        e = other["entering"].as_numpy_array()
        n = np.arange(e.size)
        p1 = pd.DataFrame(dict(zip("hklen", (*hkl, e, n))), copy=False)

        merged = pd.merge(p0, p1, on=["h", "k", "l", "e"], suffixes=[0, 1])

        n0 = cctbx.array_family.flex.size_t(merged.n0.values)
        n1 = cctbx.array_family.flex.size_t(merged.n1.values)

        return n0, n1

    @staticmethod
    def concat(
        tables: List[dials_array_family_flex_ext.reflection_table],
    ) -> dials_array_family_flex_ext.reflection_table:
        """
        Concatenate a list of reflection tables, taking care to correctly handle
        experiment identifiers and ids.
        :param tables: A list of reflection tables
        :return: A single combined reflection table
        """
        from dials.util.multi_dataset_handling import renumber_table_id_columns

        tables = renumber_table_id_columns(tables)
        first = tables[0]
        for table in tables[1:]:
            first.extend(table)
        return first

    def match_with_reference(self, other):
        """
        Match reflections with another set of reflections.

        :param other: The reflection table to match against
        :return: The matches
        """
        logger.info("Matching reference spots with predicted reflections")
        logger.info(" %d observed reflections input", len(other))
        logger.info(" %d reflections predicted", len(self))

        # Get the miller index, entering flag and turn number for
        # Both sets of reflections
        i1 = self["id"]
        h1 = self["miller_index"]
        e1 = self["entering"].as_int()
        x1, y1, z1 = self["xyzcal.px"].parts()
        p1 = self["panel"]

        i2 = other["id"]
        h2 = other["miller_index"]
        e2 = other["entering"].as_int()
        x2, y2, z2 = other["xyzcal.px"].parts()
        p2 = other["panel"]

        class Match:
            def __init__(self):
                self.a = []
                self.b = []

        # Create the match lookup
        lookup = collections.defaultdict(Match)
        for i in range(len(self)):
            item = h1[i] + (e1[i], i1[i], p1[i])
            lookup[item].a.append(i)

        # Add matches from input reflections
        for i in range(len(other)):
            item = h2[i] + (e2[i], i2[i], p2[i])
            if item in lookup:
                lookup[item].b.append(i)

        # Create the list of matches
        match1 = []
        match2 = []
        for item, value in lookup.items():
            if len(value.b) == 0:
                continue
            elif len(value.a) == 1 and len(value.b) == 1:
                match1.append(value.a[0])
                match2.append(value.b[0])
            else:
                matched = {}
                for i in value.a:
                    d = []
                    for j in value.b:
                        dx = x1[i] - x2[j]
                        dy = y1[i] - y2[j]
                        dz = z1[i] - z2[j]
                        d.append((i, j, dx**2 + dy**2 + dz**2))
                    i, j, d = min(d, key=lambda x: x[2])
                    if j not in matched:
                        matched[j] = (i, d)
                    elif d < matched[j][1]:
                        matched[j] = (i, d)
                for key1, value1 in matched.items():
                    match1.append(value1[0])
                    match2.append(key1)

        # Select everything which matches
        sind = cctbx.array_family.flex.size_t(match1)
        oind = cctbx.array_family.flex.size_t(match2)

        # Sort by self index
        sort_index = cctbx.array_family.flex.size_t(
            sorted(range(len(sind)), key=lambda x: sind[x])
        )
        sind = sind.select(sort_index)
        oind = oind.select(sort_index)

        s2 = self.select(sind)
        o2 = other.select(oind)
        h1 = s2["miller_index"]
        h2 = o2["miller_index"]
        e1 = s2["entering"]
        e2 = o2["entering"]
        assert (h1 == h2).all_eq(True)
        assert (e1 == e2).all_eq(True)
        x1, y1, z1 = s2["xyzcal.px"].parts()
        x2, y2, z2 = o2["xyzcal.px"].parts()
        distance = cctbx.array_family.flex.sqrt(
            cctbx.array_family.flex.pow2(x1 - x2)
            + cctbx.array_family.flex.pow2(y1 - y2)
            + cctbx.array_family.flex.pow2(z1 - z2)
        )
        mask = distance < 2
        logger.info(" %d reflections matched", len(o2))
        logger.info(" %d reflections accepted", mask.count(True))
        self.set_flags(sind.select(mask), self.flags.reference_spot)
        self.set_flags(sind.select(o2.get_flags(self.flags.strong)), self.flags.strong)
        self.set_flags(
            sind.select(o2.get_flags(self.flags.indexed)), self.flags.indexed
        )
        self.set_flags(
            sind.select(o2.get_flags(self.flags.used_in_refinement)),
            self.flags.used_in_refinement,
        )
        other_matched_indices = oind.select(mask)
        other_unmatched_mask = cctbx.array_family.flex.bool(len(other), True)
        other_unmatched_mask.set_selected(
            other_matched_indices,
            cctbx.array_family.flex.bool(len(other_matched_indices), False),
        )
        other_matched = other.select(other_matched_indices)
        other_unmatched = other.select(other_unmatched_mask)
        for key, column in self.select(sind.select(mask)).cols():
            other_matched[key] = column
        mask2 = cctbx.array_family.flex.bool(len(self), False)
        mask2.set_selected(sind.select(mask), True)
        return mask2, other_matched, other_unmatched

    def match(
        self,
        other: dials_array_family_flex_ext.reflection_table,
        *,
        max_separation: int = 2,
        key: str = "xyzobs.px.value",
        scale: Tuple[float, float, float] = (1, 1, 1),
    ) -> Tuple[
        cctbx.array_family.flex.int,
        cctbx.array_family.flex.int,
        cctbx.array_family.flex.double,
    ]:
        """
        Match reflections from this reflection list to another (reference) list

        The match is based on comparison of any chosen 3-vector column.

        Args:
            other: The reflection table to match against
            max_separations: Maximum distance in column-space to match on
            key: Name of 3-vector column
            scale:
                Relative scales to apply to vectors before matching in
                case e.g. very wide or very fine slicing.

        Returns:
            (self_index, other_index, distance) where:
                self_index: is the reindex in the this reflection table
                other_index: is the index in the other reflection table
                distance: The calculated distance of specified column-vector

            such that self[self_index][j] ~= other[other_index][j].
        """

        xyz = self[key]
        ref = other[key]

        if scale != (1, 1, 1):
            x, y, z = xyz.parts()
            x *= scale[0]
            y *= scale[1]
            z *= scale[2]
            xyz = cctbx.array_family.flex.vec3_double(x, y, z)

            x, y, z = ref.parts()
            x *= scale[0]
            y *= scale[1]
            z *= scale[2]
            ref = cctbx.array_family.flex.vec3_double(x, y, z)

        x = xyz.as_double().as_1d()
        r = ref.as_double().as_1d()
        ann = AnnAdaptorSelfInclude(r, 3)
        ann.query(x)

        mm = cctbx.array_family.flex.size_t_range(xyz.size())
        nn, distance = ann.nn.as_size_t(), cctbx.array_family.flex.sqrt(ann.distances)

        sel = distance <= max_separation

        mm = mm.select(sel)
        nn = nn.select(sel)
        distance = distance.select(sel)
        return mm, nn, distance

    def compute_zeta(self, experiment):
        """
        Compute zeta for each reflection.

        :param experiment: The experimental models
        :return: Zeta for each reflection
        """
        from dials.algorithms.profile_model.gaussian_rs import zeta_factor

        m2 = experiment.goniometer.get_rotation_axis()
        s0 = experiment.beam.get_s0()
        self["zeta"] = zeta_factor(m2, s0, self["s1"])
        return self["zeta"]

    def compute_zeta_multi(self, experiments):
        """
        Compute zeta for each reflection.

        :param experiments: The list of experiments
        :return: Zeta for each reflection
        """
        from dials.algorithms.profile_model.gaussian_rs import zeta_factor

        m2 = cctbx.array_family.flex.vec3_double(len(experiments))
        s0 = cctbx.array_family.flex.vec3_double(len(experiments))
        for i, e in enumerate(experiments):
            m2[i] = e.goniometer.get_rotation_axis()
            s0[i] = e.beam.get_s0()
        self["zeta"] = zeta_factor(m2, s0, self["s1"], self["id"])
        return self["zeta"]

    def compute_d_single(self, experiment):
        """
        Compute the resolution for each reflection.

        :param experiment: The experimental models
        :return: The resolution for each reflection
        """
        uc = dials_array_family_flex_ext.unit_cell(1)
        uc[0] = experiment.crystal.get_unit_cell()
        self["d"] = uc.d(
            self["miller_index"], cctbx.array_family.flex.size_t(len(self), 0)
        )
        return self["d"]

    def compute_d(self, experiments):
        """
        Compute the resolution for each reflection.

        :param experiments: The experiment list
        :return: The resolution for each reflection
        """
        uc = dials_array_family_flex_ext.unit_cell(len(experiments))
        for i, e in enumerate(experiments):
            uc[i] = e.crystal.get_unit_cell()
        assert self["id"].all_ge(0)
        self["d"] = uc.d(
            self["miller_index"], cctbx.array_family.flex.size_t(list(self["id"]))
        )
        return self["d"]

    def compute_bbox(self, experiments, sigma_b_multiplier=2.0):
        """
        Compute the bounding boxes.

        :param experiments: The list of experiments
        :param profile_model: The profile models
        :param sigma_b_multiplier: Multiplier to cover extra background
        :return: The bounding box for each reflection
        """
        self["bbox"] = dials_array_family_flex_ext.int6(len(self))
        for expr, indices in self.iterate_experiments_and_indices(experiments):
            self["bbox"].set_selected(
                indices,
                expr.profile.compute_bbox(
                    self.select(indices),
                    expr.crystal,
                    expr.beam,
                    expr.detector,
                    expr.goniometer,
                    expr.scan,
                    sigma_b_multiplier=sigma_b_multiplier,
                ),
            )
        return self["bbox"]

    def compute_partiality(self, experiments):
        """
        Compute the reflection partiality.

        :param experiments: The experiment list
        :param profile_model: The profile models
        :return: The partiality for each reflection
        """
        self["partiality"] = cctbx.array_family.flex.double(len(self))
        for expr, indices in self.iterate_experiments_and_indices(experiments):
            self["partiality"].set_selected(
                indices,
                expr.profile.compute_partiality(
                    self.select(indices),
                    expr.crystal,
                    expr.beam,
                    expr.detector,
                    expr.goniometer,
                    expr.scan,
                ),
            )
        return self["partiality"]

    def compute_mask(self, experiments, image_volume=None, overlaps=None):
        """
        Apply a mask to the shoeboxes.

        :param experiments: The list of experiments
        :param profile_model: The profile model
        """
        for expr, indices in self.iterate_experiments_and_indices(experiments):
            result = expr.profile.compute_mask(
                self.select(indices),
                expr.crystal,
                expr.beam,
                expr.detector,
                expr.goniometer,
                expr.scan,
                image_volume=image_volume,
            )
            if result is not None:
                if "fraction" not in self:
                    self["fraction"] = cctbx.array_family.flex.double(len(self))
                self["fraction"].set_selected(indices, result)

    def iterate_experiments_and_indices(self, experiments):
        """
        A helper function to iterate through experiments and indices of reflections
        for each experiment
        """
        assert len(experiments) > 0
        index_list = self.split_indices_by_experiment_id(len(experiments))
        assert len(experiments) == len(index_list)
        tot = 0
        for l in index_list:
            tot += len(l)
        assert tot == len(self)
        yield from zip(experiments, index_list)

    def random_split(self, n=2):
        """Randomly split table into n tables.

        Not all tables will be the same length."""
        n = int(n)
        if n <= 1:
            return [self]
        list_of_reflection_tables = []

        perm = cctbx.array_family.flex.random_permutation(self.size())
        divisions = [int(i * self.size() / n) for i in range(0, n + 1)]
        for i, v in enumerate(divisions[:-1]):
            isel = perm[v : divisions[i + 1]]
            if isel:
                list_of_reflection_tables.append(self.select(isel))
        return list_of_reflection_tables

    def compute_background(self, experiments, image_volume=None):
        """
        Helper function to compute the background.

        :param experiments: The list of experiments
        """
        success = self.background_algorithm(experiments).compute_background(
            self, image_volume
        )
        self.set_flags(~success, self.flags.failed_during_background_modelling)

    def compute_centroid(self, experiments, image_volume=None):
        """
        Helper function to compute the centroid.

        :param experiments: The list of experiments
        """
        self.centroid_algorithm(experiments).compute_centroid(
            self, image_volume=image_volume
        )

    def compute_summed_intensity(
        self, image_volume: dials.model.data.MultiPanelImageVolume = None
    ) -> None:
        """
        Compute intensity via summation integration.
        """
        from dials.algorithms.integration.sum import sum_integrate_and_update_table

        success = sum_integrate_and_update_table(self, image_volume=image_volume)
        self.set_flags(~success, self.flags.failed_during_summation)

    def compute_fitted_intensity(self, fitter):
        """
        Helper function to compute the intensity.

        :param experiments: The list of experiments
        :param profile_model: The profile model
        """
        success = fitter.fit(self)
        self.set_flags(~success, self.flags.failed_during_profile_fitting)

    def compute_corrections(self, experiments):
        """
        Helper function to correct the intensity.

        :param experiments: The list of experiments
        :return: The LP correction for each reflection
        """
        from dials.algorithms.integration import Corrections, CorrectionsMulti

        compute = CorrectionsMulti()
        for experiment in experiments:
            if experiment.goniometer is not None:
                compute.append(
                    Corrections(
                        experiment.beam, experiment.goniometer, experiment.detector
                    )
                )
            else:
                compute.append(Corrections(experiment.beam, experiment.detector))
        lp = compute.lp(self["id"], self["s1"])
        self["lp"] = lp
        if experiment.detector[0].get_mu() > 0:
            qe = compute.qe(self["id"], self["s1"], self["panel"])
            self["qe"] = qe
        return lp

    def extract_shoeboxes(self, imageset, mask=None, nthreads=1):
        """
        Helper function to read a load of shoebox data.

        :param imageset: The imageset
        :param mask: The mask to apply
        :param nthreads: The number of threads to use
        :return: A tuple containing read time and extract time
        """
        from time import time

        from dials.model.data import make_image

        assert "shoebox" in self
        detector = imageset.get_detector()
        try:
            frame0, frame1 = imageset.get_array_range()
        except Exception:
            frame0, frame1 = (0, len(imageset))
        extractor = dials_array_family_flex_ext.ShoeboxExtractor(
            self, len(detector), frame0, frame1
        )
        logger.info(" Beginning to read images")
        read_time = 0
        extract_time = 0
        for i in range(len(imageset)):
            logger.debug("  reading image %d", i)
            st = time()
            image = imageset.get_corrected_data(i)
            mask2 = imageset.get_mask(i)
            if mask is not None:
                assert len(mask) == len(mask2)
                mask2 = tuple(m1 & m2 for m1, m2 in zip(mask, mask2))
            read_time += time() - st
            st = time()
            extractor.next(make_image(image, mask2))
            extract_time += time() - st
            del image
        assert extractor.finished()
        logger.info("  successfully read %d images", frame1 - frame0)
        logger.info("  read time: %.1f seconds", read_time)
        logger.info("  extract time: %.1f seconds", extract_time)
        return read_time, extract_time

    def is_overloaded(self, experiments):
        """
        Check if the shoebox contains overloaded pixels.

        :param experiments: The experiment list
        :return: True/False overloaded for each reflection
        """
        from dials.algorithms.shoebox import OverloadChecker

        assert "shoebox" in self
        assert "id" in self
        detectors = [expr.detector for expr in experiments]
        checker = OverloadChecker()
        for detector in detectors:
            checker.add(
                cctbx.array_family.flex.double(
                    p.get_trusted_range()[1] for p in detector
                )
            )
        result = checker(self["id"], self["shoebox"])
        self.set_flags(result, self.flags.overloaded)
        return result

    def contains_invalid_pixels(self):
        """
        Check if the shoebox contains invalid pixels.

        :return: True/False invalid for each reflection
        """
        from dials.algorithms.shoebox import MaskCode

        assert "shoebox" in self
        x0, x1, y0, y1, z0, z1 = self["bbox"].parts()
        ntotal = (x1 - x0) * (y1 - y0) * (z1 - z0)
        assert ntotal.all_gt(0)
        sbox = self["shoebox"]
        nvalid = sbox.count_mask_values(MaskCode.Valid)
        nbackg = sbox.count_mask_values(MaskCode.Background)
        nforeg = sbox.count_mask_values(MaskCode.Foreground)
        nvalbg = sbox.count_mask_values(MaskCode.Background | MaskCode.Valid)
        nvalfg = sbox.count_mask_values(MaskCode.Foreground | MaskCode.Valid)
        ninvbg = nbackg - nvalbg
        ninvfg = nforeg - nvalfg
        assert ninvbg.all_ge(0)
        assert ninvfg.all_ge(0)
        self.set_flags(ninvbg > 0, self.flags.background_includes_bad_pixels)
        self.set_flags(ninvfg > 0, self.flags.foreground_includes_bad_pixels)
        return (ntotal - nvalid) > 0

    def find_overlaps(self, experiments=None, border=0):
        """
        Check for overlapping reflections.

        :param experiments: The experiment list
        :param tolerance: A positive integer specifying border around shoebox
        :return: The overlap list
        """
        from dials.algorithms.shoebox import OverlapFinder

        # Expand the bbox if necessary
        if border > 0:
            x0, x1, y0, y1, z0, z1 = self["bbox"].parts()
            x0 -= border
            x1 += border
            y0 -= border
            y1 += border
            z0 -= border
            z1 += border
            bbox = dials_array_family_flex_ext.int6(x0, x1, y0, y1, z0, z1)
        else:
            bbox = self["bbox"]

        # Get the panel and id
        panel = self["panel"]

        # Group according to imageset
        if experiments is not None:
            groups = itertools.groupby(
                range(len(experiments)), lambda x: experiments[x].imageset
            )

            # Get the experiment ids we're to treat together
            lookup = {}
            for j, (key, indices) in enumerate(groups):
                for i in indices:
                    lookup[i] = j
            group_id = cctbx.array_family.flex.size_t([lookup[i] for i in self["id"]])
        elif "imageset_id" in self:
            imageset_id = self["imageset_id"]
            assert imageset_id.all_ge(0)
            group_id = cctbx.array_family.flex.size_t(list(imageset_id))
        else:
            raise RuntimeError("Either need to supply experiments or have imageset_id")

        # Create the overlap finder
        find_overlapping = OverlapFinder()

        # Find the overlaps
        overlaps = find_overlapping(group_id, panel, bbox)
        assert overlaps.num_vertices() == len(self)

        # Return the overlaps
        return overlaps

    def assert_experiment_identifiers_are_consistent(self, experiments=None):
        """
        Check the experiment identifiers
        """
        identifiers = self.experiment_identifiers()
        if len(identifiers) > 0:
            values = list(identifiers.values())
            assert len(set(values)) == len(values), (len(set(values)), len(values))
            if "id" in self:
                index = set(self["id"]).difference({-1})
                for i in index:
                    assert i in identifiers, (i, list(identifiers))
        if experiments is not None:
            if len(identifiers) > 0:
                assert len(identifiers) == len(experiments), (
                    len(identifiers),
                    len(experiments),
                )
                assert len(identifiers) == len(set(experiments.identifiers()))
                for experiment in experiments:
                    assert (
                        experiment.identifier in identifiers.values()
                    ), experiment.identifier

    def are_experiment_identifiers_consistent(self, experiments=None):
        """
        Check the experiment identifiers
        """
        try:
            self.assert_experiment_identifiers_are_consistent(experiments)
        except AssertionError:
            return False
        return True

    def compute_miller_indices_in_asu(self, experiments):
        """
        Compute miller indices in the asu
        """
        self["miller_index_asu"] = cctbx.array_family.flex.miller_index(len(self))
        for idx, experiment in enumerate(experiments):

            # Create the crystal symmetry object
            uc = experiment.crystal.get_unit_cell()
            sg = experiment.crystal.get_space_group()
            cs = cctbx.crystal.symmetry(uc, space_group=sg)

            # Get the selection and compute the miller indices
            selection = self["id"] == idx
            h = self["miller_index"].select(selection)
            ms = cctbx.miller.set(cs, h)
            ms_asu = ms.map_to_asu()
            h_asu = ms_asu.indices()

            # Set the miller indices
            self["miller_index_asu"].set_selected(selection, h_asu)

        return self["miller_index_asu"]

    def select_on_experiment_identifiers(self, list_of_identifiers):
        """
        Given a list of experiment identifiers (strings), perform a selection
        and return a reflection table with properly configured experiment_identifiers
        map.
        """
        # First get the reverse of the map i.e. ids for a given exp_identifier
        id_values = []
        for k, v in zip(
            self.experiment_identifiers().keys(), self.experiment_identifiers().values()
        ):
            if v in list_of_identifiers:
                id_values.append(k)
        if len(id_values) != len(list_of_identifiers):
            logger.warning(
                """Not all requested identifiers
found in the table's map, has the experiment_identifiers() map been created?
Requested %s:
Found %s"""
                % (list_of_identifiers, id_values)
            )
        # Build up a selection and use this
        sel = cctbx.array_family.flex.bool(self.size(), False)
        for id_val in id_values:
            id_sel = self["id"] == id_val
            sel.set_selected(id_sel, True)
        self = self.select(sel)
        # Remove entries from the experiment_identifiers map
        for k in self.experiment_identifiers().keys():
            if k not in id_values:
                del self.experiment_identifiers()[k]
        return self

    def remove_on_experiment_identifiers(self, list_of_identifiers):
        """
        Remove datasets from the table, given a list of experiment
        identifiers (strings).
        """
        # First get the reverse of the map i.e. ids for a given exp_identifier
        assert "id" in self
        id_values = []
        for k, v in zip(
            self.experiment_identifiers().keys(), self.experiment_identifiers().values()
        ):
            if v in list_of_identifiers:
                id_values.append(k)
        if len(id_values) != len(list_of_identifiers):
            logger.warning(
                """Not all requested identifiers
found in the table's map, has the experiment_identifiers() map been created?
Requested %s:
Found %s"""
                % (list_of_identifiers, id_values)
            )
        # Now delete the selections, also removing the entry from the map
        for id_val in id_values:
            sel = self["id"] == id_val
            self.del_selected(sel)
            del self.experiment_identifiers()[id_val]
        return self

    def clean_experiment_identifiers_map(self):
        """
        Remove any entries from the identifier map that do not have any
        data in the table. Primarily to call as saving data to give a
        consistent table and map.
        """
        dataset_ids_in_table = set(self["id"]).difference({-1})
        dataset_ids_in_map = set(self.experiment_identifiers().keys())
        ids_to_remove = dataset_ids_in_map.difference(dataset_ids_in_table)
        for i in ids_to_remove:
            del self.experiment_identifiers()[i]

    def reset_ids(self):
        """
        Reset the 'id' column such that the experiment identifiers are
        numbered 0 .. n-1.
        """
        reverse_map = {v: k for k, v in self.experiment_identifiers()}
        orig_id = self["id"].deep_copy()
        for k in self.experiment_identifiers().keys():
            del self.experiment_identifiers()[k]
        for i_exp, exp_id in enumerate(reverse_map.keys()):
            sel_exp = orig_id == reverse_map[exp_id]
            self["id"].set_selected(sel_exp, i_exp)
            self.experiment_identifiers()[i_exp] = exp_id

    def centroid_px_to_mm(self, experiments):
        """
        Map spot centroids from pixel/image number to mm/radian.

        Used to convert spot centroids coming from e.g. dials.find_spots which are
        in pixel/image number units to mm/radian units as required for indexing and
        refinement.

        Args:
          experiments (dxtbx.model.ExperimentList): A list of experiments.
        """

        self["xyzobs.mm.value"] = cctbx.array_family.flex.vec3_double(len(self))
        self["xyzobs.mm.variance"] = cctbx.array_family.flex.vec3_double(len(self))
        # e.g. data imported from XDS; no variance known then; since is used
        # only for weights assign as 1 => uniform weights
        if "xyzobs.px.variance" not in self:
            self["xyzobs.px.variance"] = cctbx.array_family.flex.vec3_double(
                len(self), (1, 1, 1)
            )
        panel_numbers = cctbx.array_family.flex.size_t(self["panel"])

        for i, expt in enumerate(experiments):
            if "imageset_id" in self:
                sel_expt = self["imageset_id"] == i
            else:
                sel_expt = self["id"] == i
            for i_panel in range(len(expt.detector)):
                sel = sel_expt & (panel_numbers == i_panel)
                centroid_position, centroid_variance, _ = centroid_px_to_mm_panel(
                    expt.detector[i_panel],
                    expt.scan,
                    self["xyzobs.px.value"].select(sel),
                    self["xyzobs.px.variance"].select(sel),
                    cctbx.array_family.flex.vec3_double(sel.count(True), (1, 1, 1)),
                )
                self["xyzobs.mm.value"].set_selected(sel, centroid_position)
                self["xyzobs.mm.variance"].set_selected(sel, centroid_variance)

    def map_centroids_to_reciprocal_space(
        self, experiments, calculated=False, crystal_frame=False
    ):
        """Map mm/radian spot centroids to reciprocal space.

        Used to convert spot centroids provided in mm/radian units to reciprocal space
        as required for indexing. Adds the column 'rlp' to the reflection table, which
        contains a :py:class:`.flex.vec3_double` array of the reciprocal lattice vectors.

        Args:
          experiments (dxtbx.model.ExperimentList): A list of experiments.
          calculated (Bool): use calculated positions.
          crystal_frame (Bool): return x, y, z positions as divided by U matrix
        """

        self["s1"] = cctbx.array_family.flex.vec3_double(len(self))
        self["rlp"] = cctbx.array_family.flex.vec3_double(len(self))
        panel_numbers = cctbx.array_family.flex.size_t(self["panel"])

        # crystal_frame is not used in indexing, but is a feature of the
        # reciprocal lattice viewer -> but if we are looking at the crystal
        # coordinate frame we need to look at the experiments independently

        for i, expt in enumerate(experiments):
            if not crystal_frame and "imageset_id" in self:
                sel_expt = self["imageset_id"] == i
            else:
                sel_expt = self["id"] == i

            for i_panel in range(len(expt.detector)):
                sel = sel_expt & (panel_numbers == i_panel)
                if calculated:
                    x, y, rot_angle = self["xyzcal.mm"].select(sel).parts()
                else:
                    x, y, rot_angle = self["xyzobs.mm.value"].select(sel).parts()
                s1 = expt.detector[i_panel].get_lab_coord(
                    cctbx.array_family.flex.vec2_double(x, y)
                )
                s1 = s1 / s1.norms() * (1 / expt.beam.get_wavelength())
                self["s1"].set_selected(sel, s1)
                S = s1 - expt.beam.get_s0()
                if expt.goniometer is not None:
                    setting_rotation = matrix.sqr(
                        expt.goniometer.get_setting_rotation()
                    )
                    rotation_axis = expt.goniometer.get_rotation_axis_datum()
                    sample_rotation = matrix.sqr(expt.goniometer.get_fixed_rotation())
                    if expt.crystal and crystal_frame:
                        sample_rotation *= matrix.sqr(expt.crystal.get_U())

                    self["rlp"].set_selected(sel, tuple(setting_rotation.inverse()) * S)
                    self["rlp"].set_selected(
                        sel,
                        self["rlp"]
                        .select(sel)
                        .rotate_around_origin(rotation_axis, -rot_angle),
                    )
                    self["rlp"].set_selected(
                        sel, tuple(sample_rotation.inverse()) * self["rlp"].select(sel)
                    )
                else:
                    self["rlp"].set_selected(sel, S)

    def calculate_entering_flags(self, experiments):
        """Calculate the entering flags for the reflections.

        Calculate a unit vector normal to the spindle-beam plane for this experiment,
        such that the vector placed at the centre of the Ewald sphere points to the
        hemispere in which reflections cross from inside to outside of the sphere
        (reflections are exiting). Adds the array of boolean entering flags to self
        as the "entering" column.

        Note:
            NB this vector is in +ve Y direction when using imgCIF coordinate frame.

        Args:

            experiments: The experiment list to use in calculating the entering flags.
        """

        assert "s1" in self

        # Init entering flags. These are always False for experiments that have no
        # rotation axis.
        enterings = cctbx.array_family.flex.bool(len(self), False)

        for iexp, experiment in enumerate(experiments):
            if not experiment.goniometer:
                continue
            axis = matrix.col(experiment.goniometer.get_rotation_axis())
            s0 = matrix.col(experiment.beam.get_s0())
            vec = s0.cross(axis)
            sel = self["id"] == iexp
            enterings.set_selected(sel, self["s1"].dot(vec) < 0.0)

        self["entering"] = enterings

    def get(self, key, default=None):
        """
        Get item from object for given key (ex: reflection_table column).

        Returns default value if not found.
        """
        if key in self:
            return self[key]
        return default


class reflection_table_selector:
    """
    A class to select columns from reflection table.

    This is mainly useful for specifying selections from phil parameters
    """

    def __init__(self, column, op, value):
        """
        Initialise the selector

        :param col: The column name
        :param op: The operator
        :param value: The value
        """
        # Set the column and value
        self.column = column
        self.value = value

        # Set the operator
        if isinstance(op, str):
            if op == "<":
                self.op = operator.lt
            elif op == "<=":
                self.op = operator.le
            elif op == "==":
                self.op = operator.eq
            elif op == "!=":
                self.op = operator.ne
            elif op == ">=":
                self.op = operator.ge
            elif op == ">":
                self.op = operator.gt
            elif op == "&":
                self.op = operator.and_
            else:
                raise RuntimeError("Unknown operator")
        else:
            self.op = op

    @property
    def op_string(self):
        """
        Return the operator as a string
        """

        if self.op == operator.lt:
            string = "<"
        elif self.op == operator.le:
            string = "<="
        elif self.op == operator.eq:
            string = "=="
        elif self.op == operator.ne:
            string = "!="
        elif self.op == operator.ge:
            string = ">="
        elif self.op == operator.gt:
            string = ">"
        elif self.op == operator.and_:
            string = "&"
        else:
            raise RuntimeError("Unknown operator")
        return string

    def __call__(self, reflections):
        """
        Select the reflections

        :param reflections: The reflections

        :return: The selection as a mask
        """
        if self.column == "intensity.sum.i_over_sigma":
            I = reflections["intensity.sum.value"]
            V = reflections["intensity.sum.variance"]
            mask1 = V > 0
            I = I.select(mask1)
            V = V.select(mask1)
            data = I / cctbx.array_family.flex.sqrt(V)
        elif self.column == "intensity.prf.i_over_sigma":
            I = reflections["intensity.prf.value"]
            V = reflections["intensity.prf.variance"]
            mask1 = V > 0
            I = I.select(mask1)
            V = V.select(mask1)
            data = I / cctbx.array_family.flex.sqrt(V)
        else:
            mask1 = None
            data = reflections[self.column]
        if isinstance(data, cctbx.array_family.flex.double):
            value = float(self.value)
        elif isinstance(data, cctbx.array_family.flex.int):
            value = int(self.value)
        elif isinstance(data, cctbx.array_family.flex.size_t):
            value = int(self.value)
        elif isinstance(data, cctbx.array_family.flex.std_string):
            value = self.value
        elif isinstance(data, cctbx.array_family.flex.vec3_double):
            raise RuntimeError("Comparison not implemented")
        elif isinstance(data, cctbx.array_family.flex.vec2_double):
            raise RuntimeError("Comparison not implemented")
        elif isinstance(data, cctbx.array_family.flex.mat3_double):
            raise RuntimeError("Comparison not implemented")
        elif isinstance(data, dials_array_family_flex_ext.int6):
            raise RuntimeError("Comparison not implemented")
        elif isinstance(data, dials_array_family_flex_ext.shoebox):
            raise RuntimeError("Comparison not implemented")
        else:
            raise RuntimeError("Unknown column type")
        mask2 = self.op(data, value)
        if mask1 is not None:
            mask1.set_selected(
                cctbx.array_family.flex.size_t(range(len(mask1))).select(mask1), mask2
            )
        else:
            mask1 = mask2
        return mask1



from __future__ import annotations

import os
import pickle


def test_spot_counts_per_image_plot(dials_regression):
    from dials.util import ascii_art

    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "full.pickle")

    with open(pickle_path, "rb") as fh:
        refl = pickle.load(fh, encoding="bytes")
    output = ascii_art.spot_counts_per_image_plot(refl)

    expected_output = """\
116082 spots found on 540 images (max 2075 / bin)
* *      *        *                 * * *  * **            *
************************************************************
************************************************************
************************************************************
************************************************************
************************************************************
************************************************************
************************************************************
************************************************************
************************************************************
1                         image                          540"""
    output = "\n".join(line.rstrip() for line in output.splitlines())
    assert output == expected_output

    output = ascii_art.spot_counts_per_image_plot(refl, char="o", width=80, height=15)

    expected_output = """\
116082 spots found on 540 images (max 1627 / bin)
                                                         o
o oo   o  o o o       o o o o o  o  o     o o   o  o oo  oo o o   o o   o o   oo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
1                                   image                                    540"""
    output = "\n".join(line.rstrip() for line in output.splitlines())
    assert output == expected_output

    output = ascii_art.spot_counts_per_image_plot(refl, char="#", width=7, height=10)

    expected_output = """\
116082 spots found on 540 images (max 16752 / bin)
#######
#######
#######
#######
#######
#######
#######
#######
#######
#######
1   540"""
    output = "\n".join(line.rstrip() for line in output.splitlines())
    assert output == expected_output

    output = ascii_art.spot_counts_per_image_plot(
        refl.select(refl["xyzobs.px.value"].parts()[2] <= 9.5),
        char="#",
        width=10,
        height=15,
    )
    expected_output = """\
2065 spots found on 10 images (max 361 / bin)
#
#
#
#
#
#        #
#   #    #
## ##   ##
##########
##########
##########
##########
##########
##########
##########
1 image 10"""
    output = "\n".join(line.rstrip() for line in output.splitlines())
    assert output == expected_output

    # If we do spotfinding on the first 3 images, then the spot centroids will be in the
    # range z=0.5 to z=2.5
    output = ascii_art.spot_counts_per_image_plot(
        refl.select(refl["xyzobs.px.value"].parts()[2] <= 2.5),
        char="#",
        width=10,
        height=15,
    )
    expected_output = """\
624 spots found on 3 images (max 294 / bin)
#
#
#
#
#
#
##
###
###
###
###
###
###
###
###
1 3"""
    output = "\n".join(line.rstrip() for line in output.splitlines())
    assert output == expected_output



"""
Tests for the scaling model classes.
"""

from __future__ import annotations

import copy
from unittest.mock import MagicMock, Mock

import pytest

from libtbx import phil

from dials.algorithms.scaling.model.model import (
    ArrayScalingModel,
    DoseDecay,
    KBScalingModel,
    PhysicalScalingModel,
    ScalingModelBase,
    calc_n_param_from_bins,
    initialise_smooth_input,
)
from dials.array_family import flex
from dials.util.options import ArgumentParser


@pytest.fixture(scope="module")
def default_params():
    """Return the default parsed params phil scope."""
    return generated_param()


@pytest.fixture(scope="module")
def test_reflections():
    """Return a reflection table."""
    return generated_refl()


@pytest.fixture
def mock_exp():
    """Mock experiments object."""
    exp = Mock()
    exp.scan.get_oscillation.return_value = [0.0, 1.0]
    exp.scan.get_oscillation_range.return_value = [0, 90]
    exp.beam.get_s0.return_value = (0.0, 0.0, 1.01)
    exp.goniometer.get_rotation_axis.return_value = (1.0, 0.0, 0.0)
    return exp


def generated_refl():
    """Create a reflection table."""
    rt = flex.reflection_table()
    rt["xyzobs.px.value"] = flex.vec3_double([(0.1, 0.1, 0.1), (0.1, 0.1, 0.2)])
    rt["s1"] = flex.vec3_double([(0.1, 0.1, 0.1), (0.1, 0.1, 1.1)])
    rt["s1c"] = flex.vec3_double([(0.1, 0.1, 0.1), (0.1, 0.1, 1.1)])
    rt["s0c"] = flex.vec3_double([(0.0, 0.0, 1.0), (0.0, 0.0, 1.0)])
    rt["d"] = flex.double([1.0, 1.0])
    rt["batch"] = flex.int([0, 1])
    return rt


@pytest.fixture
def mock_errormodel():
    """Mock error model."""
    em = MagicMock()
    em.parameters = [1.0, 0.1]
    em.update_variances.return_value = flex.double([1.0, 1.1, 1.0, 1.0])
    return em


@pytest.fixture()
def mock_physical_params():
    """Return a mock params object for a physical model."""
    params = Mock()
    params.physical.scale_interval = 10.0
    params.physical.decay_correction = True
    params.physical.decay_interval = 15.0
    params.physical.absorption_correction = True
    params.physical.lmax = 4
    params.physical.decay_restraint = 1e-1
    params.physical.surface_weight = "auto"
    return params


def generated_param():
    """Generate the default scaling parameters object."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
  """,
        process_includes=True,
    )

    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.array.modulation_correction = True
    return parameters


def test_ScalingModelBase(mock_errormodel):
    """Test for base scaling model class"""

    configdict = {"corrections": ["1"]}
    SM_base = ScalingModelBase(configdict)
    assert not SM_base.is_scaled
    SM_base.set_scaling_model_as_scaled()
    assert SM_base.is_scaled
    SM_base.set_scaling_model_as_unscaled()
    assert not SM_base.is_scaled
    assert SM_base.configdict == configdict
    assert not SM_base.components
    _ = SM_base.to_dict()
    SM_base.set_error_model(mock_errormodel)
    assert SM_base.configdict["error_model_parameters"] == mock_errormodel.parameters
    assert SM_base.error_model is mock_errormodel
    assert "Scaling model" in str(SM_base)


def test_model_creation_from_data(default_params, mock_exp, test_reflections):
    """Test the factory creation of the three standard scaling models with the
    default params."""

    _ = KBScalingModel.from_data(default_params, [], [])

    _ = PhysicalScalingModel.from_data(default_params, mock_exp, test_reflections)

    _ = ArrayScalingModel.from_data(default_params, mock_exp, test_reflections)


def test_KBScalingModel():
    """Test for the KB Scaling Model."""

    # Test standard initialisation method.
    configdict = {"corrections": ["scale", "decay"]}
    parameters_dict = {
        "scale": {
            "parameters": flex.double([1.2]),
            "parameter_esds": flex.double([0.1]),
        },
        "decay": {
            "parameters": flex.double([0.01]),
            "parameter_esds": flex.double([0.02]),
        },
    }
    KBmodel = KBScalingModel(parameters_dict, configdict)
    assert KBmodel.id_ == "KB"
    assert "scale" in KBmodel.components
    assert "decay" in KBmodel.components
    assert list(KBmodel.components["scale"].parameters) == [1.2]
    assert list(KBmodel.components["decay"].parameters) == [0.01]
    assert list(KBmodel.components["scale"].parameter_esds) == [0.1]
    assert list(KBmodel.components["decay"].parameter_esds) == [0.02]

    # Test from_dict initialisation method.
    KB_dict = {
        "__id__": "KB",
        "scale": {
            "n_parameters": 1,
            "parameters": [0.5],
            "est_standard_devs": [0.05],
            "null_parameter_value": 1,
        },
        "configuration_parameters": {"corrections": ["scale"]},
    }
    KBmodel = KBScalingModel.from_dict(KB_dict)
    assert KBmodel.is_scaled is True
    assert "scale" in KBmodel.components
    assert "decay" not in KBmodel.components
    assert list(KBmodel.components["scale"].parameters) == [0.5]
    assert list(KBmodel.components["scale"].parameter_esds) == [0.05]

    new_dict = KBmodel.to_dict()
    assert new_dict == KB_dict

    # Test again with all parameters
    KB_dict = {
        "__id__": "KB",
        "scale": {
            "n_parameters": 1,
            "parameters": [0.5],
            "est_standard_devs": [0.05],
            "null_parameter_value": 1,
        },
        "decay": {
            "n_parameters": 1,
            "parameters": [0.2],
            "est_standard_devs": [0.02],
            "null_parameter_value": 0,
        },
        "configuration_parameters": {"corrections": ["scale", "decay"]},
    }
    KBmodel = KBScalingModel.from_dict(KB_dict)
    assert KBmodel.is_scaled is True
    assert "scale" in KBmodel.components
    assert "decay" in KBmodel.components
    assert list(KBmodel.components["scale"].parameters) == [0.5]
    assert list(KBmodel.components["scale"].parameter_esds) == [0.05]
    assert list(KBmodel.components["decay"].parameters) == [0.2]
    assert list(KBmodel.components["decay"].parameter_esds) == [0.02]

    new_dict = KBmodel.to_dict()
    assert new_dict == KB_dict

    with pytest.raises(RuntimeError):
        KB_dict["__id__"] = "physical"
        KBmodel = KBScalingModel.from_dict(KB_dict)

    assert KBmodel.consecutive_refinement_order == [["scale", "decay"]]
    assert "Decay component" in str(KBmodel)


def test_physical_model_from_data(mock_physical_params, mock_exp, test_reflections):
    """Test that it passes the correct dict to physical model."""
    physicalmodel = PhysicalScalingModel.from_data(
        mock_physical_params, mock_exp, test_reflections
    )
    assert physicalmodel.configdict["lmax"] == (mock_physical_params.physical.lmax)
    assert physicalmodel.components["absorption"].n_params == 24
    assert list(physicalmodel.components["absorption"].parameters) == [0.0] * 24

    # test updating the absorption parameters
    mock_physical_params.physical.absorption_level = "high"
    physicalmodel.update(mock_physical_params)
    assert len(physicalmodel.components["absorption"].parameters) == 48
    assert physicalmodel.configdict["abs_surface_weight"] == 5e3

    mock_physical_params.physical.absorption_level = "medium"
    physicalmodel.update(mock_physical_params)
    assert len(physicalmodel.components["absorption"].parameters) == 48
    assert physicalmodel.configdict["abs_surface_weight"] == 5e4

    mock_physical_params.physical.absorption_level = None
    mock_physical_params.physical.lmax = 4
    physicalmodel.update(mock_physical_params)
    assert len(physicalmodel.components["absorption"].parameters) == 24
    assert physicalmodel.configdict["abs_surface_weight"] == 5e4

    mock_physical_params.physical.surface_weight = 1e5
    physicalmodel.update(mock_physical_params)
    assert len(physicalmodel.components["absorption"].parameters) == 24
    assert physicalmodel.configdict["abs_surface_weight"] == 1e5

    # try fixing a parameter
    mock_physical_params.physical.correction.fix = ["decay"]
    physicalmodel = PhysicalScalingModel.from_data(
        mock_physical_params, mock_exp, test_reflections
    )
    assert physicalmodel.configdict["lmax"] == (mock_physical_params.physical.lmax)
    assert physicalmodel.components["absorption"].n_params == 24
    assert list(physicalmodel.components["absorption"].parameters) == [0.0] * 24
    assert physicalmodel.fixed_components == ["decay"]


def test_PhysicalScalingModel(test_reflections, mock_exp):
    """Test the PhysicalScalingModel class."""
    configdict = {
        "corrections": ["scale", "decay", "absorption"],
        "s_norm_fac": 1.0,
        "scale_rot_interval": 2.0,
        "d_norm_fac": 1.0,
        "decay_rot_interval": 2.0,
        "lmax": 1,
        "abs_surface_weight": 1e6,
    }

    parameters_dict = {
        "scale": {"parameters": flex.double([1.2, 1.1]), "parameter_esds": None},
        "decay": {"parameters": flex.double([0.1, 0.2]), "parameter_esds": None},
        "absorption": {
            "parameters": flex.double([0.01, 0.01, 0.01]),
            "parameter_esds": None,
        },
    }

    # Test standard factory initialisation
    physicalmodel = PhysicalScalingModel(parameters_dict, configdict)
    assert physicalmodel.id_ == "physical"
    assert "Absorption component" in str(physicalmodel)
    comps = physicalmodel.components
    assert "scale" in comps
    assert "absorption" in comps
    assert "decay" in comps
    assert list(comps["scale"].parameters) == [1.2, 1.1]
    assert list(comps["decay"].parameters) == [0.1, 0.2]
    assert list(comps["absorption"].parameters) == [0.01, 0.01, 0.01]

    # Test configure reflection table
    mock_params = Mock()
    mock_params.physical.decay_restraint = 0.0
    physicalmodel.configure_components(test_reflections, mock_exp, mock_params)

    # Test from_dict initialisation method.
    physical_dict = {
        "__id__": "physical",
        "scale": {
            "n_parameters": 2,
            "parameters": [0.5, 1.0],
            "est_standard_devs": [0.05, 0.1],
            "null_parameter_value": 1,
        },
        "configuration_parameters": {
            "corrections": ["scale"],
            "s_norm_fac": 0.1,
            "scale_rot_interval": 10.0,
            "decay_restaint": 1e-1,
        },
    }
    physicalmodel = PhysicalScalingModel.from_dict(physical_dict)
    assert physicalmodel.id_ == "physical"
    assert "scale" in physicalmodel.components
    assert "absorption" not in physicalmodel.components
    assert "decay" not in physicalmodel.components
    assert list(physicalmodel.components["scale"].parameters) == [0.5, 1.0]
    assert list(physicalmodel.components["scale"].parameter_esds) == [0.05, 0.1]

    new_dict = physicalmodel.to_dict()
    assert new_dict == physical_dict

    # Test from_dict initialisation method for all components.
    physical_dict = {
        "__id__": "physical",
        "scale": {
            "n_parameters": 2,
            "parameters": [0.5, 1.0],
            "est_standard_devs": [0.05, 0.1],
            "null_parameter_value": 1,
        },
        "decay": {
            "n_parameters": 2,
            "parameters": [0.1, 0.2],
            "est_standard_devs": [0.01, 0.01],
            "null_parameter_value": 0,
        },
        "absorption": {
            "n_parameters": 3,
            "parameters": [0.0, 0.1, 0.2],
            "est_standard_devs": [0.01, 0.02, 0.03],
            "null_parameter_value": 0,
        },
        "configuration_parameters": {
            "corrections": ["scale", "decay", "absorption"],
            "s_norm_fac": 0.1,
            "scale_rot_interval": 10.0,
            "d_norm_fac": 0.2,
            "decay_rot_interval": 20.0,
            "lmax": 1,
            "abs_surface_weight": 1e6,
        },
    }
    physicalmodel = PhysicalScalingModel.from_dict(physical_dict)
    assert physicalmodel.id_ == "physical"
    assert "scale" in physicalmodel.components
    assert "absorption" in physicalmodel.components
    assert "decay" in physicalmodel.components
    assert list(physicalmodel.components["scale"].parameters) == [0.5, 1.0]
    assert list(physicalmodel.components["scale"].parameter_esds) == [0.05, 0.1]
    assert list(physicalmodel.components["decay"].parameters) == [0.1, 0.2]
    assert list(physicalmodel.components["decay"].parameter_esds) == [0.01, 0.01]
    assert list(physicalmodel.components["absorption"].parameters) == [0.0, 0.1, 0.2]
    assert list(physicalmodel.components["absorption"].parameter_esds) == [
        0.01,
        0.02,
        0.03,
    ]

    new_dict = physicalmodel.to_dict()
    assert new_dict == physical_dict

    with pytest.raises(RuntimeError):
        physical_dict["__id__"] = "array"
        physicalmodel = PhysicalScalingModel.from_dict(physical_dict)

    assert len(physicalmodel.consecutive_refinement_order) == 2
    assert "Absorption component:" in str(physicalmodel)

    # test limit batch range
    parameters_dict = {
        "scale": {
            "n_parameters": 11,
            "parameters": [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5],
            "parameter_esds": None,
        },
        "decay": {
            "n_parameters": 11,
            "parameters": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1],
            "parameter_esds": None,
        },
    }
    configdict = {
        "corrections": ["scale", "decay"],
        "s_norm_fac": 0.1,
        "scale_rot_interval": 10.0,
        "d_norm_fac": 0.1,
        "decay_rot_interval": 10.0,
        "valid_image_range": (1, 90),
        "valid_osc_range": (0, 90),
    }
    physical = PhysicalScalingModel(parameters_dict, configdict)
    physical.limit_image_range((1, 50))
    assert list(physical.components["scale"].parameters) == [
        0.5,
        1.0,
        1.5,
        2.0,
        2.5,
        3.0,
        3.5,
    ]
    assert list(physical.components["decay"].parameters) == [
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
    ]
    assert physical.configdict["valid_osc_range"] == (0, 50)
    assert physical.configdict["valid_image_range"] == (1, 50)

    # try edge cases
    # if restricted by > rot int, then reduce number of params and shift offset
    # if necessary
    physical = PhysicalScalingModel(
        copy.deepcopy(parameters_dict), copy.deepcopy(configdict)
    )
    physical.limit_image_range((7, 45))
    assert list(physical.components["scale"].parameters) == [
        1.0,
        1.5,
        2.0,
        2.5,
        3.0,
        3.5,
    ]
    assert list(physical.components["decay"].parameters) == [
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
    ]
    assert physical.configdict["valid_osc_range"] == (6, 45)
    assert physical.configdict["valid_image_range"] == (7, 45)

    # if not restricted by > rot int, then should 'squeeze' the parameters closer
    # in rotation angle, leaving the same number of params (as reducing number of
    # params would give parameter spacing greater than initially specified rot int)
    physical = PhysicalScalingModel(
        copy.deepcopy(parameters_dict), copy.deepcopy(configdict)
    )
    physical.limit_image_range((5, 45))
    assert list(physical.components["scale"].parameters) == [
        0.5,
        1.0,
        1.5,
        2.0,
        2.5,
        3.0,
        3.5,
    ]
    assert list(physical.components["decay"].parameters) == [
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
    ]
    assert physical.configdict["valid_osc_range"] == (4, 45)
    assert physical.configdict["valid_image_range"] == (5, 45)


def test_DoseDecayModel(test_reflections, mock_exp):
    """Test the DoseDecay class."""
    configdict = {
        "corrections": ["scale", "decay", "relative_B"],
        "s_norm_fac": 1.0,
        "scale_rot_interval": 2.0,
        "resolution_dependence": "quadratic",
    }

    parameters_dict = {
        "scale": {"parameters": flex.double([1.2, 1.1]), "parameter_esds": None},
        "decay": {"parameters": flex.double([0.1]), "parameter_esds": None},
        "relative_B": {"parameters": flex.double([-0.1]), "parameter_esds": None},
    }

    def test_setup(model):
        assert model.id_ == "dose_decay"
        assert set(model.components.keys()) == {"scale", "decay", "relative_B"}
        assert list(model.components["scale"].parameters) == [1.2, 1.1]
        assert list(model.components["decay"].parameters) == [0.1]
        assert list(model.components["relative_B"].parameters) == [-0.1]

    # Test standard factory initialisation
    model = DoseDecay(parameters_dict, configdict)
    test_setup(model)

    # Test configure reflection table - adds data to components
    mock_params = Mock()
    model.configure_components(test_reflections, mock_exp, mock_params)
    assert list(model.components["decay"].data["d"]) == list(test_reflections["d"])

    d = model.to_dict()
    model_from_dict = DoseDecay.from_dict(d)
    test_setup(model_from_dict)


def test_ArrayScalingModel(test_reflections, mock_exp):
    """Test the ArrayScalingModel class."""

    configdict = {
        "corrections": ["decay", "absorption", "modulation"],
        "n_res_param": 2,
        "n_time_param": 2,
        "resmin": 1.0,
        "res_bin_width": 1.0,
        "time_norm_fac": 1.0,
        "time_rot_interval": 1.0,
        "n_x_param": 2,
        "n_y_param": 2,
        "xmin": 0.0,
        "ymin": 0.0,
        "x_bin_width": 1.0,
        "y_bin_width": 2.0,
        "n_x_mod_param": 2,
        "n_y_mod_param": 2,
        "x_det_bin_width": 2.0,
        "y_det_bin_width": 2.0,
    }

    parameters_dict = {
        "decay": {
            "parameters": flex.double([1.2, 1.1, 1.0, 0.9]),
            "parameter_esds": None,
        },
        "absorption": {
            "parameters": flex.double([0.1, 0.2, 0.1, 0.2, 0.1, 0.2, 0.1, 0.2]),
            "parameter_esds": None,
        },
        "modulation": {"parameters": flex.double(4, 0.9), "parameter_esds": None},
    }

    # Test standard factory initialisation
    arraymodel = ArrayScalingModel(parameters_dict, configdict)
    assert arraymodel.id_ == "array"
    assert "decay" in arraymodel.components
    assert "absorption" in arraymodel.components
    assert "modulation" in arraymodel.components
    assert list(arraymodel.components["decay"].parameters) == [1.2, 1.1, 1.0, 0.9]
    assert list(arraymodel.components["absorption"].parameters) == [
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
    ]
    assert list(arraymodel.components["modulation"].parameters) == 4 * [0.9]

    # Test configure reflection table
    _ = arraymodel.configure_components(test_reflections, mock_exp, [])

    # Test from_dict initialisation method for previous model case.
    init_dict = arraymodel.to_dict()
    new_array_model = ArrayScalingModel.from_dict(init_dict)
    assert new_array_model.id_ == "array"
    comps = new_array_model.components
    assert "modulation" in comps
    assert "absorption" in comps
    assert "decay" in comps
    assert list(comps["decay"].parameters) == [1.2, 1.1, 1.0, 0.9]
    assert list(comps["absorption"].parameters) == [
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
    ]
    assert list(comps["modulation"].parameters) == 4 * [0.9]

    # Test from_dict initialisation method for another case.
    array_dict = {
        "__id__": "array",
        "decay": {
            "n_parameters": 4,
            "parameters": [0.5, 1.0, 0.4, 1.0],
            "null_parameter_value": 1.0,
            "est_standard_devs": [0.05, 0.1, 0.05, 0.1],
        },
        "configuration_parameters": {
            "corrections": ["decay"],
            "n_res_param": 2,
            "n_time_param": 2,
            "resmin": 1.0,
            "res_bin_width": 1.0,
            "time_norm_fac": 1.0,
            "time_rot_interval": 1.0,
        },
    }
    arraymodel = ArrayScalingModel.from_dict(array_dict)
    assert arraymodel.id_ == "array"
    comps = arraymodel.components
    assert "modulation" not in comps
    assert "absorption" not in comps
    assert "decay" in comps
    assert list(comps["decay"].parameters) == [0.5, 1.0, 0.4, 1.0]
    assert list(comps["decay"].parameter_esds) == [0.05, 0.1, 0.05, 0.1]

    new_dict = arraymodel.to_dict()
    assert new_dict == array_dict

    with pytest.raises(RuntimeError):
        array_dict["__id__"] = "physical"
        arraymodel = ArrayScalingModel.from_dict(array_dict)

    assert len(arraymodel.consecutive_refinement_order) == 3
    assert "Decay component" in str(arraymodel)

    # test limit batch range
    configdict = {
        "corrections": ["decay", "absorption"],
        "n_res_param": 2,
        "n_time_param": 3,
        "resmin": 1.0,
        "res_bin_width": 1.0,
        "time_norm_fac": 0.1,
        "time_rot_interval": 10.0,
        "n_x_param": 2,
        "n_y_param": 2,
        "xmin": 0.0,
        "ymin": 0.0,
        "x_bin_width": 1.0,
        "y_bin_width": 2.0,
        "n_x_mod_param": 2,
        "n_y_mod_param": 2,
        "x_det_bin_width": 2.0,
        "y_det_bin_width": 2.0,
        "valid_image_range": (1, 20),
        "valid_osc_range": (0, 20),
    }

    parameters_dict = {
        "decay": {
            "parameters": flex.double([1.2, 1.1, 1.0, 0.9, 0.8, 0.7]),
            "parameter_esds": None,
        },
        "absorption": {
            "parameters": flex.double(
                [0.1, 0.2, 0.1, 0.2, 0.1, 0.2, 0.1, 0.2, 0.3, 0.4, 0.3, 0.4]
            ),
            "parameter_esds": None,
        },
    }
    array = ArrayScalingModel(parameters_dict, configdict)
    array.limit_image_range((1, 10))
    assert list(array.components["decay"].parameters) == [1.2, 1.1, 1.0, 0.9]
    assert list(array.components["absorption"].parameters) == [
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
        0.1,
        0.2,
    ]
    assert array.configdict["n_time_param"] == 2
    assert array.configdict["valid_image_range"] == (1, 10)
    assert array.configdict["valid_osc_range"] == (0, 10)


def test_model_factory_utilities():
    """Test the utility functions in the scaling_model_factory module."""

    # Test calc_n_param_from_bins(value_min, value_max, n_bins)
    assert calc_n_param_from_bins(0.0, 1.0, 1) == (2, 1.0)
    assert calc_n_param_from_bins(0.0, 2.0, 2) == (3, 1.0)
    assert calc_n_param_from_bins(0.0, 3.0, 3) == (5, 1.0)
    assert calc_n_param_from_bins(0.0, 10.0, 10) == (12, 1.0)
    assert calc_n_param_from_bins(0.0, 10.0, 5) == (7, 2.0)
    with pytest.raises(AssertionError):
        (_, _) = calc_n_param_from_bins(0.0, 1.0, 0)
    with pytest.raises(AssertionError):
        (_, _) = calc_n_param_from_bins(0.0, 1.0, 0.5)

    # Test initialise_smooth_input(osc_range, one_osc_width, interval)
    # This is initialised with the oscillation range, width of one osc and
    # rotation interval in degrees, returning
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 1.0)
    assert (n_param, norm_fac, rot_int) == (12, 0.999, 1.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 12)
    assert (n_param, norm_fac, rot_int) == (2, 0.0999, 10.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 10)
    assert (n_param, norm_fac, rot_int) == (2, 0.0999, 10.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 9.99)
    assert (n_param, norm_fac, rot_int) == (3, 0.1998, 5.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 5.0)
    assert (n_param, norm_fac, rot_int) == (3, 0.1998, 5.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 1.0, 4.99)
    assert (n_param, norm_fac, rot_int) == (5, 0.999 * 3.0 / 10.0, 10.0 / 3.0)
    n_param, norm_fac, rot_int = initialise_smooth_input([0, 10], 2.0, 4.99)
    assert (n_param, norm_fac, rot_int) == (5, 0.999 * 6.0 / 10.0, 10.0 / 3.0)



from __future__ import annotations

from cctbx import uctbx
from cctbx.miller import index_generator
from iotbx.phil import parse

from dials.array_family import flex

# The phil scope
phil_scope = parse(
    """
  powder {

    water_ice {
      unit_cell = 4.498,4.498,7.338,90,90,120
        .type = unit_cell
        .help = "The unit cell to generate d_spacings for ice rings."
      space_group = 194
        .type = space_group
        .help = "The space group used to generate d_spacings for ice rings."
      d_min = 1
        .type = float(value_min=0.0)
        .help = "The minimum resolution to filter ice rings"
      width = 0.002
        .type = float(value_min=0.0)
        .help = "The width of an ice ring (in 1/d^2)."
    }

    apply = *none water_ice
      .type = choice(multi=True)
      .help = "The power ring filters to apply"
  }
"""
)


class PowderRingFilter:
    """
    A class to do powder ring filtering.
    """

    def __init__(self, unit_cell, space_group, d_min, width):
        """
        Initialise the filter.

        :param unit_cell: The unit_cell of the powder rings
        :param space_group: The space group of the powder rings
        :param d_min: The maximum resolution to filter to
        :param width: The resolution width to filter around
        """
        assert d_min > 0
        assert width > 0

        # Correct unit cell
        unit_cell = space_group.average_unit_cell(unit_cell)

        self.half_width = width / 2.0
        d_min = uctbx.d_star_sq_as_d(uctbx.d_as_d_star_sq(d_min) + self.half_width)

        # Generate a load of indices
        generator = index_generator(unit_cell, space_group.type(), False, d_min)
        indices = generator.to_array()

        # Compute d spacings and sort by resolution
        self.d_star_sq = flex.sorted(unit_cell.d_star_sq(indices))

    def __call__(self, d):
        """
        True if within powder ring.

        :param d: The resolution
        :return: True/False in powder ring
        """
        result = flex.bool(len(d), False)
        d_star_sq = uctbx.d_as_d_star_sq(d)
        for ds2 in self.d_star_sq:
            result = result | (flex.abs(d_star_sq - ds2) < self.half_width)
        return result


class IceRingFilter:
    """
    A class to do ice ring filtering
    """

    def __init__(self):
        """
        Initialise the filter.

        :param width: The resolution width to filter around
        """
        # Hexagonal ice ring resolution ranges in 1/d^2
        self.ice_rings = [
            (0.0640, 0.0690),
            (0.0710, 0.0780),
            (0.0825, 0.0880),
            (0.138, 0.144),
            (0.190, 0.205),
            (0.228, 0.240),
            (0.262, 0.266),
            (0.267, 0.278),
            (0.280, 0.288),
            (0.337, 0.341),
            (0.429, 0.435),
            (0.459, 0.466),
            (0.478, 0.486),
            (0.531, 0.537),
        ]

    def __call__(self, d):
        """
        True if within powder ring.

        :param d: The resolution
        :return: True/False in powder ring
        """
        result = flex.bool(len(d), False)
        d2 = 1.0 / flex.pow2(d)
        for ice_ring in self.ice_rings:
            result = result | (d2 >= ice_ring[0]) & (d2 <= ice_ring[1])
        return result



"""
Classes to initialise a 'parameter manager', to indicate to a
refiner which components of the model are to be refined.
"""

from __future__ import annotations

import logging

from scitbx import sparse

from dials.array_family import flex

logger = logging.getLogger("dials")


class active_parameter_manager:
    """
    Class to manage the current active parameters during minimisation.

    The parameter manager is initialised with components - a dict of ('name' : obj)
    pairs, and a selection list, which is a list of 'name' values to be minimised
    in this cycle. obj is the component of a model, and this code requires that
    obj has the attribute 'parameters'.
    This class stores a reference to the components to be minimised, alongside
    some bookkeeping to allow selection of the parameters for an individual
    component.
    """

    def __init__(self, components, selection_list):
        # self.target = target
        self.x = flex.double([])
        self.components = {}
        self.derivatives = None
        self.var_cov_matrix = None
        self.components_list = []  # just a list of the component names
        n_cumul_params = 0
        for component, obj in components.items():
            if component in selection_list:
                assert hasattr(
                    obj, "parameters"
                ), """component object must have the
          attribute 'parameters' for access to the component parameters."""
                self.x.extend(obj.free_parameters)
                n_params = len(obj.free_parameters)
                self.components.update(
                    {
                        component: {
                            "object": obj,
                            "n_params": n_params,
                            "start_idx": n_cumul_params,
                            "end_idx": len(self.x),
                        }
                    }
                )
                n_cumul_params += n_params
        self.n_active_params = len(self.x)
        for comp in self.components:
            self.components_list.extend([comp])

    def select_parameters(self, component):
        """Select the subset of self.x corresponding to the component (a string)."""
        start_idx = self.components[component]["start_idx"]
        end_idx = self.components[component]["end_idx"]
        return self.x[start_idx:end_idx]

    def set_param_vals(self, x):
        """Set method for refinement engine access."""
        self.x = x
        for component in self.components:
            component_obj = self.components[component]["object"]
            component_obj.free_parameters = self.select_parameters(component)

    def get_param_vals(self):
        """Get method for refinement engine access."""
        return self.x

    def get_param_names(self):
        l = []
        for name, data in self.components.items():
            l.extend([name + f"_param{j}" for j in range(data["n_params"])])
        return l

    def calculate_model_state_uncertainties(self, var_cov):
        """Set var_cov matrices for each component to allow later calculation
        of errors."""
        i = 0
        for component in self.components.values():
            n = component["n_params"]
            sub = var_cov.matrix_copy_block(i, i, n, n)
            component["object"].var_cov_matrix = sub
            i += n
        self.var_cov_matrix = var_cov

    def set_param_esds(self, esds):
        """Set the estimated standard deviations of the model components."""
        for component in self.components.values():
            start_idx = component["start_idx"]
            end_idx = component["end_idx"]
            component["object"].free_parameter_esds = esds[start_idx:end_idx]


class TargetInterface:
    def compute_functional_gradients(self, block):
        return self.target.compute_functional_gradients(block)

    def compute_restraints_functional_gradients(self, block):
        return self.target.compute_restraints_functional_gradients(block)

    def compute_residuals_and_gradients(self, block):
        return self.target.compute_residuals_and_gradients(block)

    def compute_restraints_residuals_and_gradients(self, block):
        return self.target.compute_restraints_residuals_and_gradients(block)

    def compute_residuals(self, block):
        return self.target.compute_residuals(block)


class multi_active_parameter_manager(TargetInterface):
    """
    Parameter manager to manage the current active parameters during minimisation
    for multiple datasets that are being minimised simultaneously.

    Initialise with two lists of components and selections, each item of which
    is used to initialise an active parameter manager of type apm_class.
    """

    def __init__(self, target, components_list, selection_lists, apm_class):
        self.target = target
        self.x = flex.double([])
        self.derivatives = None
        self.components_list = []  # A list of the component names.
        self.apm_list = []
        self.apm_data = {}
        all_same_components = False
        if all(i == selection_lists[0] for i in selection_lists[1:]):
            logger.info(
                "Components to be refined in this cycle for all datasets: %s",
                "".join(str(i) + ", " for i in selection_lists[0]).rstrip(", "),
            )
            all_same_components = True
        for j, (components, selection_list) in enumerate(
            zip(components_list, selection_lists)
        ):
            self.apm_list.append(apm_class(components, selection_list))
            if not all_same_components:
                logger.info(
                    "Components to be refined in this cycle for dataset %s: %s",
                    j,
                    ",".join(i for i in self.apm_list[j].components_list),
                )
        n_cumul_params = 0
        for i, apm in enumerate(self.apm_list):
            self.x.extend(apm.x)
            n_params = apm.n_active_params
            self.apm_data.update(
                {i: {"start_idx": n_cumul_params, "end_idx": len(self.x)}}
            )
            n_cumul_params += n_params
        self.n_active_params = len(self.x)

        for apm in self.apm_list:
            for comp in apm.components:
                self.components_list.extend([comp])

    def select_parameters(self, apm_number):
        """Select the subset of self.x corresponding to the apm number."""
        apm_data = self.apm_data[apm_number]
        return self.x[apm_data["start_idx"] : apm_data["end_idx"]]

    def set_param_vals(self, x):
        """Set method for refinement engine access."""
        self.x = x
        for i, single_apm in enumerate(self.apm_list):
            single_apm.set_param_vals(self.select_parameters(i))

    def get_param_vals(self):
        """Get method for refinement engine access."""
        return self.x

    def get_param_names(self):
        l = []
        for i, apm in enumerate(self.apm_list):
            for name, data in apm.components.items():
                l.extend([name + f"_expt{i}_param{j}" for j in range(data["n_params"])])
        return l

    def set_param_esds(self, esds):
        """Set the estimated standard deviations of the parameters."""
        for i, apm in enumerate(self.apm_list):
            apm_data = self.apm_data[i]
            apm.set_param_esds(esds[apm_data["start_idx"] : apm_data["end_idx"]])

    def calculate_model_state_uncertainties(self, var_cov):
        """Set var_cov matrices for each component, to allow later calculation
        of errors."""
        i = 0
        for apm in self.apm_list:
            n = apm.n_active_params
            sub = var_cov.matrix_copy_block(i, i, n, n)
            apm.calculate_model_state_uncertainties(sub)
            i += n


class shared_active_parameter_manager(multi_active_parameter_manager):

    """Class to enforce sharing of model components.

    Intercept calls to a multi_apm, to override set_params calls and manage
    reshaping of gradient/jacobian during minimisation.
    """

    def __init__(self, target, components_list, selection_lists, apm_class, shared):
        super().__init__(target, components_list, selection_lists, apm_class)
        n_unique_params = 0
        found_initial_shared = False
        # first loop over to work out how many unique parameters overall
        for i, apm in enumerate(self.apm_list):
            for name, comp in apm.components.items():
                if name != shared:
                    n_unique_params += comp["n_params"]
                elif not found_initial_shared:
                    n_unique_params += comp["n_params"]
                    found_initial_shared = True

        self.reducing_matrix = sparse.matrix(self.n_active_params, n_unique_params)
        # now loop through to work out reducing matrix
        # also need to update apm_data with a size_t selection
        found_initial_shared = False
        shared_params = (0, 0)
        cumul_params = 0
        unique_parameters = []
        for i, apm in enumerate(self.apm_list):
            apm_sel = flex.size_t()
            # ^ needs to be size_t to make sure selected in correct order
            for name, comp in apm.components.items():
                start_col_idx = cumul_params
                indiv_start = self.apm_data[i]["start_idx"] + comp["start_idx"]
                indiv_end = self.apm_data[i]["start_idx"] + comp["end_idx"]
                if name != shared:
                    apm_sel.extend(
                        flex.size_t(
                            range(cumul_params, cumul_params + comp["n_params"])
                        )
                    )
                    for n, j in enumerate(range(indiv_start, indiv_end)):
                        self.reducing_matrix[j, start_col_idx + n] = 1
                        unique_parameters.append(j)
                    cumul_params += comp["n_params"]  #
                elif not found_initial_shared:
                    apm_sel.extend(
                        flex.size_t(
                            range(cumul_params, cumul_params + comp["n_params"])
                        )
                    )
                    for n, j in enumerate(range(indiv_start, indiv_end)):
                        self.reducing_matrix[j, start_col_idx + n] = 1
                        unique_parameters.append(j)
                    shared_start_column = start_col_idx
                    found_initial_shared = True
                    shared_params = (indiv_start, indiv_end)
                    cumul_params += comp["n_params"]
                else:  # name == shared and n_shared_found > 0:
                    assert (
                        shared_params[1] - shared_params[0] == indiv_end - indiv_start
                    )
                    apm_sel.extend(
                        flex.size_t(range(shared_params[0], shared_params[1]))
                    )
                    for n, j in enumerate(range(indiv_start, indiv_end)):
                        self.reducing_matrix[j, shared_start_column + n] = 1
            self.apm_data[i]["apm_sel"] = apm_sel

        # also need to reduce self.x to the size of the new params
        self.x = self.x.select(flex.size_t(unique_parameters))

    def select_parameters(self, apm_number):
        """Select the subset of self.x corresponding to the apm number."""
        apm_data = self.apm_data[apm_number]
        sel = apm_data["apm_sel"]
        return self.x.select(sel)

    def compute_functional_gradients(self, block):
        f, g = self.target.compute_functional_gradients(block)
        return f, g * self.reducing_matrix

    def compute_restraints_functional_gradients(self, block):
        res = self.target.compute_restraints_functional_gradients(block)
        if res is not None:
            return res[0], res[1] * self.reducing_matrix
        return res

    def compute_residuals_and_gradients(self, block):
        r, j, w = self.target.compute_residuals_and_gradients(block)
        return r, j * self.reducing_matrix, w

    def compute_restraints_residuals_and_gradients(self, block):
        res = self.target.compute_restraints_residuals_and_gradients(block)
        if res is not None:
            return res[0], res[1] * self.reducing_matrix, res[2]
        return res

    def set_param_esds(self, esds):
        """Set the estimated standard deviations of the parameters."""
        for apm, apm_data in zip(self.apm_list, self.apm_data.values()):
            apm.set_param_esds(esds.select(apm_data["apm_sel"]))

    def calculate_model_state_uncertainties(self, var_cov):
        """Set var_cov matrices for each component, to allow later calculation
        of errors."""
        for i, apm in enumerate(self.apm_list):
            sub_var_cov = sparse.matrix(apm.n_active_params, apm.n_active_params)
            n_this = 0
            for comp in apm.components.values():
                n = comp["n_params"]
                start_idx = self.apm_data[i]["apm_sel"][n_this]
                sub = var_cov.matrix_copy_block(start_idx, start_idx, n, n)
                sub_var_cov.assign_block(sub, n_this, n_this)
                n_this += n
            apm.calculate_model_state_uncertainties(sub_var_cov.as_dense_matrix())


class ParameterManagerGenerator:
    """
    Class to generate multi-dataset parameter managers for minimisation.

    Handles the case of concurrent parameter minimisation (one parameter
    manager generated) and the case of consecutive parameter minimisation
    (several parameter managers generated, one for each minimisation, depending
    on the data_manager.consecutive_refinement_order property).
    """

    def __init__(self, data_managers, apm_type, target, mode="concurrent", shared=None):
        if mode not in ["concurrent", "consecutive"]:
            raise ValueError(
                "Bad value for refinement order mode: %s, expected %s"
                % (mode, " or ".join(["concurrent", "consecutive"]))
            )
        self.target = target
        self.data_managers = data_managers
        self.apm_type = apm_type
        self.mode = mode
        self.shared = shared
        self.param_lists = [None] * len(data_managers)
        if self.mode == "concurrent":
            for i, data_manager in enumerate(self.data_managers):
                components = list(data_manager.components)
                for f in data_manager.fixed_components:
                    try:
                        components.remove(f)
                    except ValueError:
                        continue
                self.param_lists[i] = components
        else:  # mode=consecutive
            # Generate nested list indicating the names of active parameters
            # e.g consecutive_order for class is [["a", "b"], ["c"]],
            # data_man has components ['a'], return [["a"]]
            # data_man has components ['a', 'c'], return [["a"], ["c"]]
            for i, data_manager in enumerate(self.data_managers):
                ind_param_list = []
                for cycle in data_manager.consecutive_refinement_order:
                    corrlist = [
                        corr for corr in cycle if corr in data_manager.components
                    ]
                    for f in data_manager.fixed_components:
                        try:
                            corrlist.remove(f)
                        except ValueError:
                            continue
                    if corrlist:
                        ind_param_list.append(corrlist)
                self.param_lists[i] = ind_param_list

    def parameter_managers(self):
        """Generate the parameter managers for each cycle of refinement."""
        if self.mode == "concurrent":
            return self._parameter_managers_concurrent()
        return self._parameter_managers_consecutive()

    def _parameter_managers_concurrent(self):
        components = [s.components for s in self.data_managers]
        if self.shared:
            return [
                shared_active_parameter_manager(
                    self.target,
                    components,
                    self.param_lists,
                    self.apm_type,
                    self.shared,
                )
            ]
        return [
            multi_active_parameter_manager(
                self.target, components, self.param_lists, self.apm_type
            )
        ]

    def _parameter_managers_consecutive(self):
        components = [s.components for s in self.data_managers]
        while any(self.param_lists[i] for i, _ in enumerate(self.data_managers)):
            params = []
            for param_list in self.param_lists:
                if param_list:
                    params.append(param_list.pop(0))
                else:
                    params.append([])
            if self.shared:
                yield shared_active_parameter_manager(
                    self.target, components, params, self.apm_type, self.shared
                )
            else:
                yield multi_active_parameter_manager(
                    self.target, components, params, self.apm_type
                )



from __future__ import annotations

import logging
import math

from cctbx import crystal, uctbx, xray
from libtbx import libtbx, phil
from scitbx import fftpack, matrix
from scitbx.array_family import flex

import dials_algorithms_indexing_ext
from dials.algorithms import indexing

from .strategy import Strategy
from .utils import group_vectors, is_approximate_integer_multiple

logger = logging.getLogger(__name__)


fft3d_phil_str = """\
b_iso = Auto
    .type = float(value_min=0)
    .expert_level = 2
rmsd_cutoff = 15
    .type = float(value_min=0)
    .expert_level = 1
peak_search = *flood_fill clean
    .type = choice
    .expert_level = 2
peak_volume_cutoff = 0.15
    .type = float
    .expert_level = 2
reciprocal_space_grid {
    n_points = 256
        .type = int(value_min=0)
        .expert_level = 1
    d_min = Auto
        .type = float(value_min=0)
        .help = "The high resolution limit in Angstrom for spots to include in "
                "the initial indexing."
    }
"""


class FFT3D(Strategy):
    """
    Basis vector search using a 3D FFT in reciprocal space.

    Reciprocal space is sampled as a 3D Cartesian grid, aligned with the basis of the
    laboratory frame. The reciprocal-space positions of the centroids of measured
    spots are ascribed a value of 1, with the rest of the grid assigned a value of 0.
    A 3D FFT is performed and the three shortest non-collinear reciprocal spatial
    wave vectors with appreciable spectral weight correspond to the basis vectors of
    the real space lattice.

    Because this procedure requires a sampling of all of reciprocal space, up to the
    d* value of the measured spot with the highest resolution, it can be more memory
    intensive than alternative approaches. To mitigate this, the 3D FFT will
    sometimes be curtailed to a region of reciprocal space below a certain
    resolution, and higher-resolution spots will be ignored.

    See:
        Bricogne, G. (1986). Proceedings of the EEC Cooperative Workshop on Position-Sensitive Detector Software (Phase III), p. 28. Paris: LURE.
        Campbell, J. W. (1998). J. Appl. Cryst. 31, 407-413.
    """

    phil_help = (
        "Search for the basis vectors of the direct lattice by performing a 3D FFT in "
        "reciprocal space of the density of found spots. Since this can be quite "
        "memory-intensive, the data used for indexing may automatically be "
        "constrained to just the lower resolution spots."
    )

    phil_scope = phil.parse(fft3d_phil_str)

    def __init__(self, max_cell, min_cell=3, params=None, *args, **kwargs):
        """Construct an FFT3D object.

        Args:
            max_cell (float): An estimate of the maximum cell dimension of the primitive
                cell.
            n_points (int): The size of the fft3d grid.
            d_min (float): The high resolution limit in Angstrom for spots to include in
                the initial indexing. If `Auto` then calculated as
                `d_min = 5 * max_cell/n_points`.
            b_iso (float): Apply an isotropic b_factor weight to the points when doing
                the FFT. If `Auto` then calculated as
                `b_iso = -4 * d_min ** 2 * math.log(0.05)`.
            rmsd_cutoff (float): RMSD cutoff applied to the transformed real-space map
                prior to performing the peak search.
            peak_volume_cutoff (float): Only include peaks that are larger than this
                fraction of the volume of the largest peak in the transformed real-space
                map.
            min_cell (float): A conservative lower bound on the minimum possible
                primitive unit cell dimension.
        """
        super().__init__(max_cell, params=params, *args, **kwargs)
        n_points = self._params.reciprocal_space_grid.n_points
        self._gridding = fftpack.adjust_gridding_triple(
            (n_points, n_points, n_points), max_prime=5
        )
        self._n_points = self._gridding[0]
        self._min_cell = min_cell

    def find_basis_vectors(self, reciprocal_lattice_vectors):
        """Find a list of likely basis vectors.

        Args:
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors to search for periodicity.

        Returns:
            A tuple containing the list of basis vectors and a flex.bool array
            identifying which reflections were used in indexing.
        """
        if self._params.reciprocal_space_grid.d_min is libtbx.Auto:
            # rough calculation of suitable d_min based on max cell
            # see also Campbell, J. (1998). J. Appl. Cryst., 31(3), 407-413.
            # fft_cell should be greater than twice max_cell, so say:
            #   fft_cell = 2.5 * max_cell
            # then:
            #   fft_cell = n_points * d_min/2
            #   2.5 * max_cell = n_points * d_min/2
            # a little bit of rearrangement:
            #   d_min = 5 * max_cell/n_points

            max_cell = self._max_cell
            d_min = 5 * max_cell / self._n_points
            d_spacings = 1 / reciprocal_lattice_vectors.norms()
            d_min = max(d_min, min(d_spacings))
            logger.info("Setting d_min: %.2f", d_min)
        else:
            d_min = self._params.reciprocal_space_grid.d_min

        grid_real, used_in_indexing = self._fft(reciprocal_lattice_vectors, d_min)
        self.sites, self.volumes = self._find_peaks(grid_real, d_min)

        # hijack the xray.structure class to facilitate calculation of distances

        self.crystal_symmetry = crystal.symmetry(
            unit_cell=self._fft_cell, space_group_symbol="P1"
        )
        xs = xray.structure(crystal_symmetry=self.crystal_symmetry)
        for i, site in enumerate(self.sites):
            xs.add_scatterer(xray.scatterer("C%i" % i, site=site))

        xs = xs.sites_mod_short()
        sites_cart = xs.sites_cart()
        lengths = flex.double([matrix.col(sc).length() for sc in sites_cart])
        perm = flex.sort_permutation(lengths)
        xs = xs.select(perm)
        volumes = self.volumes.select(perm)

        vectors = xs.sites_cart()
        norms = vectors.norms()
        sel = (norms > self._min_cell) & (norms < (2 * self._max_cell))
        vectors = vectors.select(sel)
        vectors = [matrix.col(v) for v in vectors]
        volumes = volumes.select(sel)

        vector_groups = group_vectors(vectors, volumes)
        vectors = [g.mean for g in vector_groups]
        volumes = flex.double(max(g.weights) for g in vector_groups)

        # sort by peak size
        perm = flex.sort_permutation(volumes, reverse=True)
        volumes = volumes.select(perm)
        vectors = [vectors[i] for i in perm]

        for i, (v, volume) in enumerate(zip(vectors, volumes)):
            logger.debug(f"{i} {v.length()} {volume}")

        # sort by length
        lengths = flex.double(v.length() for v in vectors)
        perm = flex.sort_permutation(lengths)

        # exclude vectors that are (approximately) integer multiples of a shorter
        # vector
        unique_vectors = []
        unique_volumes = flex.double()
        for p in perm:
            v = vectors[p]
            is_unique = True
            for i, v_u in enumerate(unique_vectors):
                if (unique_volumes[i] > volumes[p]) and is_approximate_integer_multiple(
                    v_u, v
                ):
                    logger.debug(
                        "rejecting %s: integer multiple of %s", v.length(), v_u.length()
                    )
                    is_unique = False
                    break
            if is_unique:
                unique_vectors.append(v)
                unique_volumes.append(volumes[p])

        # re-sort by peak volume
        perm = flex.sort_permutation(unique_volumes, reverse=True)
        self.candidate_basis_vectors = [unique_vectors[i] for i in perm]
        return self.candidate_basis_vectors, used_in_indexing

    def _fft(self, reciprocal_lattice_vectors, d_min):

        (
            reciprocal_space_grid,
            used_in_indexing,
        ) = self._map_centroids_to_reciprocal_space_grid(
            reciprocal_lattice_vectors, d_min
        )

        logger.info(
            "Number of centroids used: %i", (reciprocal_space_grid > 0).count(True)
        )

        # gb_to_bytes = 1073741824
        # bytes_to_gb = 1/gb_to_bytes
        # (128**3)*8*2*bytes_to_gb
        # 0.03125
        # (256**3)*8*2*bytes_to_gb
        # 0.25
        # (512**3)*8*2*bytes_to_gb
        # 2.0

        fft = fftpack.complex_to_complex_3d(self._gridding)
        grid_complex = flex.complex_double(
            reals=reciprocal_space_grid,
            imags=flex.double(reciprocal_space_grid.size(), 0),
        )
        grid_transformed = fft.forward(grid_complex)
        grid_real = flex.pow2(flex.real(grid_transformed))
        del grid_transformed

        return grid_real, used_in_indexing

    def _map_centroids_to_reciprocal_space_grid(
        self, reciprocal_lattice_vectors, d_min
    ):
        logger.info("FFT gridding: (%i,%i,%i)" % self._gridding)

        grid = flex.double(flex.grid(self._gridding), 0)

        if self._params.b_iso is libtbx.Auto:
            self._params.b_iso = -4 * d_min**2 * math.log(0.05)
            logger.debug("Setting b_iso = %.1f", self._params.b_iso)
        used_in_indexing = flex.bool(reciprocal_lattice_vectors.size(), True)
        dials_algorithms_indexing_ext.map_centroids_to_reciprocal_space_grid(
            grid,
            reciprocal_lattice_vectors,
            used_in_indexing,  # do we really need this?
            d_min,
            b_iso=self._params.b_iso,
        )
        return grid, used_in_indexing

    def _find_peaks(self, grid_real, d_min):
        grid_real_binary = grid_real.deep_copy()
        rmsd = math.sqrt(
            flex.mean(
                flex.pow2(
                    grid_real_binary.as_1d() - flex.mean(grid_real_binary.as_1d())
                )
            )
        )
        grid_real_binary.set_selected(
            grid_real_binary < (self._params.rmsd_cutoff) * rmsd, 0
        )
        grid_real_binary.as_1d().set_selected(grid_real_binary.as_1d() > 0, 1)
        grid_real_binary = grid_real_binary.iround()
        from cctbx import masks

        # real space FFT grid dimensions
        cell_lengths = [self._n_points * d_min / 2 for i in range(3)]
        self._fft_cell = uctbx.unit_cell(cell_lengths + [90] * 3)

        flood_fill = masks.flood_fill(grid_real_binary, self._fft_cell)
        if flood_fill.n_voids() < 4:
            # Require at least peak at origin and one peak for each basis vector
            raise indexing.DialsIndexError(
                "Indexing failed: fft3d peak search failed to find sufficient number of peaks."
            )

        # the peak at the origin might have a significantly larger volume than the
        # rest so exclude any anomalously large peaks from determining minimum volume
        from scitbx.math import five_number_summary

        outliers = flex.bool(flood_fill.n_voids(), False)
        grid_points_per_void = flood_fill.grid_points_per_void()
        min_x, q1_x, med_x, q3_x, max_x = five_number_summary(grid_points_per_void)
        iqr_multiplier = 5
        iqr_x = q3_x - q1_x
        cut_x = iqr_multiplier * iqr_x
        outliers.set_selected(grid_points_per_void.as_double() > (q3_x + cut_x), True)
        # print q3_x + cut_x, outliers.count(True)
        isel = (
            grid_points_per_void
            > int(
                self._params.peak_volume_cutoff
                * flex.max(grid_points_per_void.select(~outliers))
            )
        ).iselection()

        sites = flood_fill.centres_of_mass_frac().select(isel)
        volumes = flood_fill.grid_points_per_void().select(isel)
        return sites, volumes



from __future__ import annotations

import os


def test(dials_regression):
    import numpy as np

    from iotbx.xds import integrate_hkl, xparm
    from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter

    from dials.algorithms.spot_prediction import IndexGenerator
    from dials.util import ioutil

    # The XDS files to read from
    integrate_filename = os.path.join(
        dials_regression, "data", "sim_mx", "INTEGRATE.HKL"
    )
    gxparm_filename = os.path.join(dials_regression, "data", "sim_mx", "GXPARM.XDS")

    # Read the XDS files
    integrate_handle = integrate_hkl.reader()
    integrate_handle.read_file(integrate_filename)
    gxparm_handle = xparm.reader()
    gxparm_handle.read_file(gxparm_filename)

    # Get the parameters we need from the GXPARM file
    d_min = 1.6
    space_group_type = ioutil.get_space_group_type_from_xparm(gxparm_handle)
    cfc = coordinate_frame_converter(gxparm_filename)
    unit_cell = cfc.get_unit_cell()

    # Generate the indices
    index_generator = IndexGenerator(unit_cell, space_group_type, d_min)
    miller_indices = index_generator.to_array()

    # Get individual generated hkl
    gen_h = [hkl[0] for hkl in miller_indices]
    gen_k = [hkl[1] for hkl in miller_indices]
    gen_l = [hkl[2] for hkl in miller_indices]

    # Get individual xds generated hkl
    xds_h = [hkl[0] for hkl in integrate_handle.hkl]
    xds_k = [hkl[1] for hkl in integrate_handle.hkl]
    xds_l = [hkl[2] for hkl in integrate_handle.hkl]

    # Get min/max generated hkl
    min_gen_h, max_gen_h = np.min(gen_h), np.max(gen_h)
    min_gen_k, max_gen_k = np.min(gen_k), np.max(gen_k)
    min_gen_l, max_gen_l = np.min(gen_l), np.max(gen_l)

    # Get min/max xds generated hkl
    min_xds_h, max_xds_h = np.min(xds_h), np.max(xds_h)
    min_xds_k, max_xds_k = np.min(xds_k), np.max(xds_k)
    min_xds_l, max_xds_l = np.min(xds_l), np.max(xds_l)

    # Ensure we have the whole xds range  in the generated set
    assert min_gen_h <= min_xds_h and max_gen_h >= max_xds_h
    assert min_gen_k <= min_xds_k and max_gen_k >= max_xds_k
    assert min_gen_l <= min_xds_l and max_gen_l >= max_xds_l



from __future__ import annotations

from os import path

import procrunner
import pytest

from dxtbx.serialize import load

from dials.command_line.modify_geometry import phil_scope, update


def test_run(dials_regression, tmp_path):
    orig_expt_json = path.join(
        dials_regression, "experiment_test_data", "kappa_experiments.json"
    )
    assert path.exists(orig_expt_json)

    orig_expt = load.experiment_list(orig_expt_json, check_format=False)
    orig_gonio = orig_expt.goniometers()[0]
    assert orig_gonio.get_angles() == pytest.approx([0, 180, 0])

    result = procrunner.run(
        ["dials.modify_geometry", orig_expt_json, "angles=10,20,30"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    new_expt_json = tmp_path / "modified.expt"
    assert new_expt_json.is_file()

    new_expt = load.experiment_list(new_expt_json, check_format=False)

    new_gonio = new_expt.goniometers()[0]
    assert new_gonio.get_angles() == pytest.approx([10, 20, 30])


def test_update(dials_data):

    orig_expt = dials_data("aluminium_standard", pathlib=True) / "imported.expt"
    assert orig_expt.is_file()

    orig_expt = load.experiment_list(orig_expt, check_format=False)
    orig_beam = orig_expt.beams()[0]
    assert orig_beam.get_wavelength() == pytest.approx(0.02508235604)

    working_params = phil_scope.fetch().extract()
    working_params.geometry.beam.wavelength = 0.05

    new_expt = update(orig_expt, working_params)

    new_beam = new_expt.beams()[0]
    assert new_beam.get_wavelength() == pytest.approx(0.05)



from __future__ import annotations


def test_gltbx_is_available():
    """
    This is not a real test for dials.rlv, which is slightly difficult to write.
    However, one common error mode is that the gltbx libraries are not available
    because they were not built earlier. This will reliably cause dials.rlv to
    fail even thought the build setup was apparently fine.
    """
    import gltbx.gl

    assert gltbx.gl.ext



from __future__ import annotations

import procrunner

import iotbx.mtz
from cctbx import uctbx
from libtbx.test_utils import approx_equal


def test(dials_data, tmpdir):
    g = sorted(f for f in dials_data("x4wide", pathlib=True).glob("*"))
    assert len(g) == 90

    commands = [
        ["dials.import"] + g,
        ["dials.slice_sequence", "imported.expt", "image_range=80,90"],
        ["dials.find_spots", "imported_80_90.expt", "nproc=1"],
        ["dials.index", "imported_80_90.expt", "strong.refl", "space_group=P41212"],
        ["dials.refine", "indexed.expt", "indexed.refl", "scan_varying=True"],
        ["dials.integrate", "refined.expt", "indexed.refl", "nproc=1"],
        [
            "dials.export",
            "refined.expt",
            "integrated.refl",
            "partiality_threshold=0.99",
        ],
    ]

    for cmd in commands:
        # print cmd
        result = procrunner.run(cmd, working_directory=tmpdir)
        assert not result.returncode and not result.stderr

    integrated_mtz = tmpdir.join("integrated.mtz")
    assert integrated_mtz.check(file=1)

    mtz_object = iotbx.mtz.object(file_name=integrated_mtz.strpath)
    assert mtz_object.column_labels()[:14] == [
        "H",
        "K",
        "L",
        "M_ISYM",
        "BATCH",
        "IPR",
        "SIGIPR",
        "I",
        "SIGI",
        "BG",
        "SIGBG",
        "FRACTIONCALC",
        "XDET",
        "YDET",
    ]

    assert len(mtz_object.batches()) == 11
    batch = mtz_object.batches()[0]
    expected_unit_cell = uctbx.unit_cell((42.5787, 42.5787, 40.2983, 90, 90, 90))
    assert expected_unit_cell.is_similar_to(uctbx.unit_cell(list(batch.cell()))), (
        expected_unit_cell.parameters(),
        list(batch.cell()),
    )
    assert mtz_object.space_group().type().lookup_symbol() == "P 41 21 2"
    assert approx_equal(mtz_object.n_reflections(), 7446, eps=2e3)



from __future__ import annotations

import logging

import gemmi
import numpy as np

from dxtbx.model.experiment_list import ExperimentList
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix
from scitbx.math import r3_rotation_axis_and_angle_from_matrix

from dials.algorithms.refinement.rotation_decomposition import (
    solve_r3_rotation_for_angles_given_axes,
)
from dials.array_family import flex
from dials.command_line.frame_orientations import extract_experiment_data

logger = logging.getLogger(__name__)


def rotate_crystal(crystal, Rmat, axis, angle):

    Amats = []
    if crystal.num_scan_points > 0:
        scan_pts = list(range(crystal.num_scan_points))
        Amats = [
            Rmat
            * matrix.sqr(crystal.get_U_at_scan_point(t))
            * matrix.sqr(crystal.get_B_at_scan_point(t))
            for t in scan_pts
        ]

    crystal.rotate_around_origin(axis, angle, deg=False)
    if Amats:
        crystal.set_A_at_scan_points(Amats)

    return crystal


class PETSOutput:
    """Class to export integrated data in PETS CIF format"""

    def __init__(self, experiments, reflections, params):
        """Init with the parameters, experiments and reflections"""

        self.filename_prefix = params.pets.filename_prefix
        self.exp_id = params.pets.id
        self.partiality_cutoff = params.pets.partiality_cutoff
        self.flag_filtering = params.pets.flag_filtering
        self.excitation_error_cutoff = params.pets.virtual_frame.excitation_error_cutoff
        self.n_merged = params.pets.virtual_frame.n_merged
        self.step = params.pets.virtual_frame.step

        # Resolve a single intensity choice
        if "sum" in params.intensity and "profile" in params.intensity:
            logger.info(
                "For format=pets, intensity must be either 'sum' or 'profile'. Defaulting to 'sum'"
            )
            self.intensity_column = "intensity.sum.value"
            self.intensity_variance_column = "intensity.sum.variance"
        elif "sum" in params.intensity:
            if len(params.intensity) > 1:
                logger.info("Setting intensity=sum")
            self.intensity_column = "intensity.sum.value"
            self.intensity_variance_column = "intensity.sum.variance"
        elif "profile" in params.intensity:
            if len(params.intensity) > 1:
                logger.info("Setting intensity=profile")
            self.intensity_column = "intensity.prf.value"
            self.intensity_variance_column = "intensity.prf.variance"
        else:
            raise ValueError("The choice intensity= must be either 'sum' or 'profile'")

        self.experiments = experiments
        if len(reflections) != 1:
            raise ValueError("Only a single reflection file can be exported")
        self.reflections = reflections[0]

        self._check_experiments()
        self._check_reflections()
        self._reorient_coordinate_frame()

        # Calculate the frame orientation data
        self.frame_orientations = extract_experiment_data(self.experiment, scale=1)

        # Calculate the rotated missetting matrices. The orientations give the
        # full SRFMU matrix for each frame, where U is the scan-static
        # orientation matrix and M is the scan-varying misset. We want the
        # SRFM part, so post-multiply by the transpose of U
        Ut = matrix.sqr(self.experiment.crystal.get_U()).transpose()
        self.frame_orientations["rotated_missets"] = [
            M * Ut for M in self.frame_orientations["orientations"]
        ]

    @staticmethod
    def _rescale_zone_axis(axis):
        """Scale a vector so that the element with largest magnitude has
        magnitude 1.0, and also invert the vector. This is so the frame
        orientations zone axes match PETS format"""
        return -axis / max(axis.each_abs().elems)

    def _check_experiments(self):
        """Extract a single experiment from an experiment list and check that
        it is suitable for cif_pets output"""

        if self.exp_id is None:
            if len(self.experiments) != 1:
                raise ValueError("Please select a single experiment for export")
            else:
                self.exp_id = 0

        try:
            self.experiment = self.experiments[self.exp_id]
        except IndexError as e:
            raise IndexError(
                f"Unable to select experiment {self.exp_id} for export"
            ) from e

        # Check the unit cell is static
        crystal = self.experiment.crystal
        if crystal.num_scan_points > 0:
            Bmat = crystal.get_B()
            if not all(
                np.allclose(crystal.get_B_at_scan_point(i), Bmat)
                for i in range(crystal.num_scan_points)
            ):
                raise ValueError(
                    "The crystal must not have a scan-varying unit cell for export"
                )

        # Check no other models are scan-varying
        if (
            self.experiment.goniometer.num_scan_points > 0
            or self.experiment.beam.num_scan_points > 0
        ):
            raise ValueError("Only scan-varying crystal orientation is supported")

    def _check_reflections(self):
        """Check and filter the reflection table to include only the reflections
        for output"""

        self.reflections = self.reflections.select(
            self.reflections["id"] == self.exp_id
        )
        self.reflections["id"] *= 0

        required_keys = (self.intensity_column, self.intensity_variance_column)
        if any(x not in self.reflections for x in required_keys):
            raise ValueError(
                f"The reflection table requires these keys: {', '.join(required_keys)}"
            )

        # Select only fully-recorded reflections
        fulls = self.reflections["partiality"] >= self.partiality_cutoff
        logger.info(f"Removing {fulls.count(False)} partial reflections")
        self.reflections = self.reflections.select(fulls)

        # Select only "integrated" reflections, if requested
        if self.flag_filtering:
            if required_keys[0] == "intensity.sum.value":
                filter_flag = self.reflections.flags.integrated_sum
            else:
                filter_flag = self.reflections.flags.integrated_prf
            self.reflections = self.reflections.select(
                self.reflections.get_flags(filter_flag)
            )

    def _reorient_coordinate_frame(self):
        """Align a DIALS experiment and data in a reflection table to the
        PETS coordinate system. In that system, the s0 vector is aligned with
        -Z, while the rotation axis is in the X-Z plane, close to +X"""

        axis = matrix.col(self.experiment.goniometer.get_rotation_axis())
        us0 = matrix.col(self.experiment.beam.get_unit_s0())

        normal = us0.cross(-axis).normalize()
        orthogonalised_axis = us0.cross(normal).normalize()

        # Keep track of s1.us0 values to check reorientation
        s1_dot_us0 = self.reflections["s1"].dot(us0)

        R = align_reference_frame(us0, (0, 0, -1), orthogonalised_axis, (1, 0, 0))

        axis_angle = r3_rotation_axis_and_angle_from_matrix(R)
        axis = axis_angle.axis
        angle = axis_angle.angle(deg=False)
        logger.info(
            "Rotating experiment about axis ({:.4f}, {:.4f}, {:.4f}) by {:.3f}°".format(
                *axis, np.degrees(angle)
            )
        )

        self.experiment.detector.rotate_around_origin(axis, angle, deg=False)
        self.experiment.crystal = rotate_crystal(
            self.experiment.crystal, R, axis, angle
        )

        # Following does not work (https://github.com/cctbx/dxtbx/issues/454)
        # self.experiment.beam.rotate_around_origin(axis, angle, deg=False)
        # Set unit s0 and polarization normal directly instead (preserving inconsistent
        # beam, if that's what we have).
        new_us0 = (R * matrix.col(self.experiment.beam.get_unit_s0())).normalize()
        new_p_norm = (
            R * matrix.col(self.experiment.beam.get_polarization_normal())
        ).normalize()
        self.experiment.beam.set_unit_s0(new_us0)
        self.experiment.beam.set_polarization_normal(new_p_norm)

        # Rotating the goniometer is also complicated. See https://github.com/cctbx/dxtbx/pull/451
        new_datum = (
            R * matrix.col(self.experiment.goniometer.get_rotation_axis_datum())
        ).normalize()
        new_F = (
            R
            * matrix.sqr(self.experiment.goniometer.get_fixed_rotation())
            * R.transpose()
        )
        new_S = (
            R
            * matrix.sqr(self.experiment.goniometer.get_setting_rotation())
            * R.transpose()
        )
        self.experiment.goniometer.set_rotation_axis_datum(new_datum)
        self.experiment.goniometer.set_setting_rotation(new_S)
        self.experiment.goniometer.set_fixed_rotation(new_F)

        # Re-calculate s1 vectors and reciprocal lattice points with new geometry
        el = ExperimentList()
        el.append(self.experiment)
        self.reflections.map_centroids_to_reciprocal_space(el, calculated=True)

        error = flex.abs(s1_dot_us0 - self.reflections["s1"].dot(new_us0))
        if flex.max(error) > 1e-10:
            raise RuntimeError("Failed to rotate experiment correctly")

    def _set_virtual_frames(self):
        """Create a list of virtual frames for the experiment, each containing
        a table of reflections and the orientation information for the virtual
        frame"""

        # Get the frame orientation data
        # directions = self.frame_orientations["directions"]
        zone_axes = self.frame_orientations["zone_axes"]
        # real_space_axes = self.frame_orientations["real_space_axes"]
        rotated_missets = self.frame_orientations["rotated_missets"]

        # Get experiment geometry
        scan = self.experiment.scan
        s0 = matrix.col(self.experiment.beam.get_s0())
        inv_wl = s0.length()
        A = matrix.sqr(self.experiment.crystal.get_A())

        _, _, frames = self.reflections["xyzcal.px"].parts()
        scan = self.experiment.scan
        self.virtual_frames = []
        arr_start, arr_end = scan.get_array_range()
        # Loop over the virtual frames
        starts = list(range(arr_start, arr_end, self.step))
        for start in starts:
            stop = start + self.n_merged
            if stop > arr_end:
                break

            # Look up the orientation data using an index, which is the centre
            # of the virtual frame, offset so that the scan starts from 0
            centre = (start + stop) / 2.0
            index = int(centre) - arr_start
            M = rotated_missets[index]
            za = zone_axes[index]

            # If the number of merged frames is even, then we need to average
            # this orientation data with that from the previous frame to get the
            # orientation in the centre of the virtual frame
            if centre - int(centre) == 0.0:
                Mprev = rotated_missets[index - 1]
                RR = M * Mprev.transpose()
                (
                    angle,
                    axis,
                ) = RR.r3_rotation_matrix_as_unit_quaternion().unit_quaternion_as_axis_and_angle(
                    deg=False
                )
                R = axis.axis_and_angle_as_r3_rotation_matrix(angle / 2, deg=False)
                M = R * Mprev

                za_prev = zone_axes[index - 1]
                delta = za - za_prev
                za = za_prev + delta / 2

            # Rescale zone axis to match PETS format
            za = self._rescale_zone_axis(za)

            # Calculate the excitation error at this orientation
            UB = flex.mat3_double(len(self.reflections), M * A)
            r = UB * self.reflections["miller_index"].as_vec3_double()
            s1 = r + s0
            excitation_err = inv_wl - s1.norms()

            # Select only the reflections within the excitation error cutoff
            # and sort
            refs = self.reflections.select(
                abs(excitation_err) <= self.excitation_error_cutoff
            )
            refs.sort("miller_index")

            self.virtual_frames.append(
                {
                    "reflections": refs,
                    "rotated_misset": M,
                    "zone_axis": za,
                }
            )

    def write_dyn_cif_pets(self):

        self._set_virtual_frames()

        cif_filename = self.filename_prefix + ".cif_pets"
        cell_params = self.experiment.crystal.get_unit_cell().parameters()
        volume = self.experiment.crystal.get_unit_cell().volume()
        wavelength = self.experiment.beam.get_wavelength()
        UBmatrix = matrix.sqr(self.experiment.crystal.get_A()).as_numpy_array()
        # tilt semi-angle of the virtual frame
        precession_angle = (
            self.experiment.scan.get_oscillation()[1] * self.n_merged
        ) / 2

        # Some of these values are left as dummy zeroes for DIALS
        comment = f""";
data collection geometry: continuous rotation
dstarmax:  0.000
RC width:  0.00000
mosaicity:  0.000
rotation axis position:  000.000
reflection size:    0.000
Virtual frame settings: number of merged frames:  {self.n_merged}
                        step between frames:      {self.step}
                        sum all intensities:      1
;"""
        uvws = []
        for frame_id, virtual_frame in enumerate(self.virtual_frames):
            u, v, w = virtual_frame["zone_axis"]
            R = virtual_frame["rotated_misset"]
            # Decompose U = Rω * Rα * Rβ, where:
            # α is around 1,0,0
            # β is around 0,1,0
            # ω is around 0,0,1
            omega, alpha, beta = solve_r3_rotation_for_angles_given_axes(
                R, (0, 0, 1), (1, 0, 0), (0, 1, 0), deg=True
            )
            scale = 1  # dummy value
            uvws += [
                [frame_id + 1, u, v, w, precession_angle, alpha, beta, omega, scale]
            ]

        ref_list = []
        for frame_id, virtual_frame in enumerate(self.virtual_frames):
            refs = virtual_frame["reflections"]
            refs["intensity.sigma"] = flex.sqrt(refs[self.intensity_variance_column])
            for r in refs.rows():
                h, k, l = r["miller_index"]
                i = r[self.intensity_column]
                sig_i = r["intensity.sigma"]
                ref_list.append([h, k, l, i, sig_i, frame_id + 1])

        self._write_dyn_cif_pets(
            cif_filename,
            cell_params,
            volume,
            wavelength,
            UBmatrix,
            uvws,
            ref_list,
            comment,
        )

    def _write_dyn_cif_pets(
        self, name, lat_params, vol, lam, UBmatrix, uvws, ref_list, comment
    ):
        """
        - name : name of dataset
        - lat_params : (a,b,c,alpha,beta,gamma)
        - vol : volume
        - lam : wavelength
        - UBmatrix : 3x3 matrix orientation
        - uvws : list of lists with elements corresponding to the diffrn_zone_axis loop
        - ref_list : list of lists with elements corresponding to the refln_ loop
        """
        doc = gemmi.cif.Document()
        b = doc.add_new_block("pets")

        pair = [
            "_cell_length_a",
            "_cell_length_b",
            "_cell_length_c",
            "_cell_angle_alpha",
            "_cell_angle_beta",
            "_cell_angle_gamma",
            "_cell_volume",
            "_diffrn_radiation_wavelength",
            "_diffrn_orient_matrix_UB_11",
            "_diffrn_orient_matrix_UB_12",
            "_diffrn_orient_matrix_UB_13",
            "_diffrn_orient_matrix_UB_21",
            "_diffrn_orient_matrix_UB_22",
            "_diffrn_orient_matrix_UB_23",
            "_diffrn_orient_matrix_UB_31",
            "_diffrn_orient_matrix_UB_32",
            "_diffrn_orient_matrix_UB_33",
            "_diffrn_pets_omega",
        ]
        values = np.hstack([lat_params, [vol, lam], UBmatrix.flatten(), [0]])
        for k, v in zip(pair, values):
            b.set_pair(k, "%.5f" % v)
        b.set_pair("_diffrn_measurement_details", comment)

        #### orientation info
        cols = [
            "id",
            "u",
            "v",
            "w",
            "precession_angle",
            "alpha",
            "beta",
            "omega",
            "scale",
        ]
        lo = b.init_loop("_diffrn_zone_axis_", cols)
        row_fmt = "{:4}," + "{:8.5f}," * 3 + "{:12.3f}," + "{:9.3f}," * 3 + "{:9.3f}"
        for uvw in uvws:
            lo.add_row(row_fmt.format(*uvw).split(","))

        #### Intensities
        logger.info("Writing intensities to cif")
        colsI = [
            "index_h",
            "index_k",
            "index_l",
            "intensity_meas",
            "intensity_sigma",
            "zone_axis_id",
        ]
        li = b.init_loop("_refln_", colsI)
        row_fmt = " " + "{:3}," * 3 + "{:11.2f}," * 2 + "{:5}"
        for row in ref_list:
            li.add_row(row_fmt.format(*row).split(","))

        doc.write_file(name)
        logger.info(f"File {name} written")
        return



"""
A docstring

This can double as a helpful message which explains how the program is run.
"""

from __future__ import annotations

import logging
import sys

# We need to parse command-line arguments to PHIL scopes.
import libtbx.phil

# Information about the experiment geometry and meta-data are recorded in
# dxtbx.model.Experiment objects, collated in ExperimentList objects.
from dxtbx.model import ExperimentList

# All command-line DIALS programs should run with dials.util.show_mail_handle_errors.
import dials.util

# The logging module is used to raise log messages.  Additionally, dials.util.log
# sets up handlers and filters for a consistent logging style across DIALS.
import dials.util.log

# Often, we deal with flex arrays and reflection tables to handle reflection data.
from dials.array_family import flex

# The DIALS option parser is based on the (old) standard Python option parser,
# but contains customisations such as the parsing of PHIL parameters.
# flatten_experiments & flatten_reflections are useful for combining multiple input
# experiment lists and reflection tables into a single instance of each.
from dials.util.options import ArgumentParser, flatten_experiments, flatten_reflections

# Useful to know what version of DIALS we are running
from dials.util.version import dials_version

try:
    from typing import List
except ImportError:
    pass


# Define a logger.
logger = logging.getLogger("dials.boilerplate")

# Define the master PHIL scope for this program.
phil_scope = libtbx.phil.parse(
    """
    output {
        reflections = stronger.refl
            .type = path
            .help = "Help strings are helpful, wherever possible."
        log = dials.command_name.log
            .type = path
    }
    bool_parameter = False
      .type = bool
    integer_parameter = 0
      .type = int
    """
)


def do_boilerplate(
    experiments: ExperimentList,
    reflections: flex.reflection_table,
    params: libtbx.phil.scope_extract,
) -> (ExperimentList, flex.reflection_table):
    """
    Write the behaviour of the program as functions and classes outside run().

    Don't include file output here, remember that this function may be re-used
    elsewhere by someone who doesn't need the output written immediately to file.

    It can be especially helpful to document any expected exceptions that might be
    raised, in order to keep track of what needs to be handled in any code that
    re-uses this function.

    Args:
        experiments:  An experiment list.
        reflections:  A reflection table.
        params:       Some parameters, in the form of a scope_extract object,
                      which is the usable form of a parsed PHIL scope.

    Raises:
        RuntimeError:  if someone says 'goose'.
    """
    logger.info("Hello world!")

    # Here's an example of an error that might be raised, as documented above.
    if "goose" in ["duck", "duck", "duck"]:
        raise RuntimeError("Quick, run!")

    logger.info("The input reflection table contains %d reflections.", len(reflections))
    logger.info(
        "The input experiment list contains %d imagesets.", len(experiments.imagesets())
    )

    logger.info("integer_parameter: %i", params.integer_parameter)
    logger.info("bool_parameter: %s", params.bool_parameter)

    return experiments, reflections


@dials.util.show_mail_on_error()
def run(args: List[str] = None, phil: libtbx.phil.scope = phil_scope) -> None:
    """
    Check command-line input and call other functions to do the legwork.

    Run the script, parsing arguments found in 'args' and using the PHIL scope
    defined in 'phil'.

    Try to keep this function minimal, defining only what is necessary to run
    the program from the command line.

    Args:
        args: The arguments supplied by the user (default: sys.argv[1:])
        phil: The PHIL scope definition (default: phil_scope, the master PHIL scope
        for this program).
    """
    usage = "dials.command_name [options] imported.expt strong.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging.
    dials.util.log.config(options.verbose, logfile=params.output.log)

    # Log the dials version
    logger.info(dials_version())

    # Log the difference between the PHIL scope definition and the active PHIL scope,
    # which will include the parsed user inputs.
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    # These functions are commonly used to collate the input.
    experiments = flatten_experiments(params.input.experiments)
    reflections = flatten_reflections(params.input.reflections)

    # You might well wish to check here that the command-line input is appropriate.
    if len(reflections) != 1:
        sys.exit("Exactly one reflection file needed.")
    if len(experiments) != 1:
        sys.exit("Exactly one experiment list required.")

    # Do whatever this program is supposed to do.
    experiments, reflections = do_boilerplate(experiments, reflections[0], params)

    # Do the file output here.
    logger.info("Writing the reflection table to %s", params.output.reflections)
    reflections.as_file(params.output.reflections)


# Keep this minimal. Calling run() should do exactly the same thing as running this
if __name__ == "__main__":
    run()



from __future__ import annotations

from scitbx.array_family import flex

from dials.algorithms.image.filter import convolve
from dials.algorithms.statistics import BinnedStatistics

# Module-level definition imported by the image viewer
phil_str = """
n_iqr = 6
    .type = int
    .help = "IQR multiplier for determining the threshold value"

blur = narrow wide
    .type = choice
    .help = "Optional preprocessing of the image by a convolution with"
            "a simple Gaussian kernel of size either 3×3 (narrow) or"
            "5×5 (wide). This may help to reduce noise peaks and to"
            "combine split spots."

n_bins = 100
    .type = int
    .help = "Number of 2θ bins in which to calculate background"
"""


class RadialProfileSpotFinderThresholdExt:
    """
    Extension to calculate a radial profile threshold. This method calculates
    background value and iqr in 2θ shells, then sets a threshold at a level
    n_iqr above the radial background. As such, it is important to have the
    beam centre correct and to mask out any significant shadows. The method may
    be particularly useful for electron diffraction images, where there can be
    considerable inelastic scatter around low resolution spots. In addition, the
    algorithm is relatively insensitive to noise properties of the detector.
    This helps for the case of integrating detectors with poorly known gain
    and response statistics.

    A similar algorithm is available in other programs. The description of
    'peakfinder 8' in https://doi.org/10.1107/S1600576714007626 was helpful
    in the development of this method.
    """

    name = "radial_profile"

    @staticmethod
    def phil():
        from libtbx.phil import parse

        phil = parse(phil_str)
        return phil

    def __init__(self, params):
        """
        Initialise the algorithm.

        :param params: The input parameters
        """
        self.params = params

        # Set approximate Gaussian kernel for blurring
        if self.params.spotfinder.threshold.radial_profile.blur == "narrow":
            # fmt: off
            self.kernel = flex.double(
                (0.0625, 0.125, 0.0625,
                 0.125,  0.25,  0.125,
                 0.0625, 0.125, 0.0625)
            )
            # fmt: on
            self.kernel.reshape(flex.grid((3, 3)))
        elif self.params.spotfinder.threshold.radial_profile.blur == "wide":
            # fmt: off
            self.kernel = (
                flex.double(
                    (
                        1,  4,  7,  4,  1,
                        4, 16, 26, 16,  4,
                        7, 26, 41, 26,  7,
                        4, 16, 26, 16,  4,
                        1,  4,  7,  4,  1,
                    )
                ) / 273
            )
            # fmt: on
            self.kernel.reshape(flex.grid((5, 5)))
        else:
            self.kernel = None

    def compute_threshold(
        self, image, mask, *, imageset, i_panel, region_of_interest=None, **kwargs
    ):
        r"""
        Compute the threshold.

        :param image: The image to process
        :param mask: The pixel mask on the image
        :\*\*kwargs: Arbitrary keyword arguments
        :returns: A boolean mask showing foreground/background pixels
        """

        if self.kernel:
            image = convolve(image, self.kernel)

        panel = imageset.get_detector()[i_panel]
        beam = imageset.get_beam()

        # Get 2θ array for the panel or ROI
        two_theta_array = panel.get_two_theta_array(beam.get_s0())
        if region_of_interest:
            x0, x1, y0, y1 = region_of_interest
            two_theta_array = two_theta_array[y0:y1, x0:x1]

        # Convert to 2θ bin selections
        lookup = two_theta_array - flex.min(two_theta_array)
        n_bins = self.params.spotfinder.threshold.radial_profile.n_bins
        multiplier = n_bins / flex.max(lookup + 1e-10)
        lookup *= multiplier  # values now in range [0,n_bins+1)
        lookup = (
            flex.floor(lookup).iround().as_size_t()
        )  # values now in range [0,n_bins-1]

        # Calculate median intensity and IQR within each bin of masked values
        masked_lookup = lookup.select(mask.as_1d())
        masked_image = image.select(mask.as_1d())
        binned_statistics = BinnedStatistics(masked_image, masked_lookup, n_bins)
        med_I = binned_statistics.get_medians()
        iqr = binned_statistics.get_iqrs()

        # Determine the threshold value for each bin. This should be at least
        # 1 quantum greater value than the median to avoid selecting everything
        # in low background cases
        n_iqr = self.params.spotfinder.threshold.radial_profile.n_iqr
        add_level = n_iqr * iqr
        adu = 1 / panel.get_gain()
        add_level.set_selected(add_level <= adu, 2.0 * adu)
        threshold = med_I + add_level

        # Now construct a threshold image
        thresh_im = threshold.select(lookup.as_1d())

        # Peaks are unmasked pixels greater than the threshold
        peaks = image > thresh_im
        return peaks & mask



from __future__ import annotations

import random
from collections import namedtuple

import pytest

from dxtbx.serialize import load
from scitbx import matrix

from dials.algorithms.profile_model.gaussian_rs.transform import beam_vector_map

SequenceAndModel = namedtuple(
    "SequenceAndModel", ("sequence", "beam", "detector", "gonio", "scan")
)


@pytest.fixture
def sequence_and_model(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )
    return SequenceAndModel(
        sequence=sequence,
        beam=sequence.get_beam(),
        detector=sequence.get_detector(),
        gonio=sequence.get_goniometer(),
        scan=sequence.get_scan(),
    )


def test_at_corners(sequence_and_model):
    assert len(sequence_and_model.detector) == 1

    # The detector beam vectors
    ds1 = beam_vector_map(sequence_and_model.detector[0], sequence_and_model.beam, True)
    expected_size = sequence_and_model.detector[0].get_image_size()[::-1]
    expected_size = tuple(e + 1 for e in expected_size)
    assert ds1.all() == expected_size

    s0_length = matrix.col(sequence_and_model.beam.get_s0()).length()

    # Ensure a few random points are correct
    eps = 1e-7
    for k in range(1000):
        j = random.randint(0, ds1.all()[0] - 1)
        i = random.randint(0, ds1.all()[1] - 1)
        y = float(j)
        x = float(i)
        xyz = sequence_and_model.detector[0].get_pixel_lab_coord((x, y))
        s11 = matrix.col(xyz).normalize() * s0_length
        s12 = matrix.col(ds1[j, i])
        assert (s11 - s12).length() <= eps


def test_sub_division_at_corners(sequence_and_model):
    n_div = 2

    # The detector beam vectors
    ds1 = beam_vector_map(
        sequence_and_model.detector[0],
        sequence_and_model.beam,
        n_div,
        True,
    )
    expected_size = sequence_and_model.detector[0].get_image_size()[::-1]
    expected_size = tuple(e * n_div + 1 for e in expected_size)
    assert ds1.all() == expected_size

    s0_length = matrix.col(sequence_and_model.beam.get_s0()).length()

    # Ensure a few random points are correct
    eps = 1e-7
    for k in range(1000):
        j = random.randint(0, ds1.all()[0] - 1)
        i = random.randint(0, ds1.all()[1] - 1)
        y = float(j) / n_div
        x = float(i) / n_div
        xyz = sequence_and_model.detector[0].get_pixel_lab_coord((x, y))
        s11 = matrix.col(xyz).normalize() * s0_length
        s12 = matrix.col(ds1[j, i])
        assert (s11 - s12).length() <= eps


def test_sub_division_at_centres(sequence_and_model):
    n_div = 2

    # The detector beam vectors
    ds1 = beam_vector_map(
        sequence_and_model.detector[0],
        sequence_and_model.beam,
        n_div,
        False,
    )
    expected_size = sequence_and_model.detector[0].get_image_size()[::-1]
    expected_size = tuple(e * n_div for e in expected_size)
    assert ds1.all() == expected_size

    s0_length = matrix.col(sequence_and_model.beam.get_s0()).length()

    # Ensure a few random points are correct
    eps = 1e-7
    for k in range(1000):
        j = random.randint(0, ds1.all()[0] - 1)
        i = random.randint(0, ds1.all()[1] - 1)
        y = float(j + 0.5) / n_div
        x = float(i + 0.5) / n_div
        xyz = sequence_and_model.detector[0].get_pixel_lab_coord((x, y))
        s11 = matrix.col(xyz).normalize() * s0_length
        s12 = matrix.col(ds1[j, i])
        assert (s11 - s12).length() <= eps



"""
Handle dynamic aspects of setup.py and building.

This is separate because the non-dynamic stuff can generally be moved
out of a setup.py, but mainly because at the moment it's how poetry
offloads the unresolved build phases.
"""
from __future__ import annotations

import ast
import itertools
import re
import sys
from pathlib import Path
from typing import Any


def get_entry_point(filename: Path, prefix: str, import_path: str) -> list[str]:
    """Returns the entry point string for a given path.

    This looks for LIBTBX_SET_DISPATCHER_NAME, and a root function
    named 'run'. It can return multiple results for each file, if more
    than one dispatcher name is bound.

    Args:
        filename:
            The python file to parse. Will look for a run() function
            and any number of LIBTBX_SET_DISPATCHER_NAME.
        prefix: The prefix to output the entry point console script with
        import_path: The import path to get to the package the file is in

    Returns:
        A list of entry_point specifications
    """
    contents = filename.read_text(encoding="utf-8")
    tree = ast.parse(contents)
    # Find root functions named "run"
    has_run = any(
        x for x in tree.body if isinstance(x, ast.FunctionDef) and x.name == "run"
    )
    if not has_run:
        return []
    # Find if we need an alternate name via LIBTBX_SET_DISPATCHER_NAME
    alternate_names = re.findall(
        r"^#\s*LIBTBX_SET_DISPATCHER_NAME\s+(.*)$", contents, re.M
    )
    if alternate_names:
        return [f"{name}={import_path}.{filename.stem}:run" for name in alternate_names]

    return [f"{prefix}.{filename.stem}={import_path}.{filename.stem}:run"]


def build(setup_kwargs: dict[str, Any]) -> None:
    """Called by setup.py to inject any dynamic configuration"""
    package_path = Path(__file__).parent / "src" / "dials"
    entry_points = setup_kwargs.setdefault("entry_points", {})
    console_scripts = entry_points.setdefault("console_scripts", [])
    # Work out what dispatchers to add
    all_dispatchers = sorted(
        itertools.chain.from_iterable(
            get_entry_point(f, "dials", "dials.command_line")
            for f in (package_path / "command_line").glob("*.py")
        )
    )
    console_scripts.extend(x for x in all_dispatchers if x not in console_scripts)
    libtbx_dispatchers = entry_points.setdefault("libtbx.dispatcher.script", [])
    libtbx_dispatchers.extend(
        "{name}={name}".format(name=x.split("=")[0]) for x in console_scripts
    )

    print(f"Found {len(entry_points['console_scripts'])} dials dispatchers")


if __name__ == "__main__":
    sys.exit("Cannot call build.py directly, please use setup.py instead")
